"""Base model classes.

- Model: Base class for an LLM model, containing most common functionality.
- TextModel: Base class for a model with text-based input/output.
- FormattedTextModel: Base class for a model that uses formatted text (chat templates) for input/output.
- MessagesModel: Base class for a model with message-based input/output.
- Tokenizer: Base tokenizer class to encode and decode tokens, measure text length in tokens, track special tokens.
"""

from typing import Any, Optional, Union, Callable, Annotated, Literal, get_origin, get_args
from abc import ABC, abstractmethod
from enum import IntEnum

import sys, json

from copy import copy
from pprint import pformat

import logging
logger = logging.getLogger(__name__)

from dataclasses import dataclass, is_dataclass

from jsonschema import ( # type: ignore[import-untyped]
    validate as json_schema_validate, 
    ValidationError as json_schema_ValidationError,
    SchemaError as json_schema_SchemaError
)

import jinja2
from jinja2.exceptions import TemplateError
from jinja2.sandbox import ImmutableSandboxedEnvironment
version = jinja2.__version__.split('.')
if int(version[0]) < 3:
    raise ImportError

def jinja_raise_exception(message):
    raise TemplateError(message)            

from pydantic import BaseModel


from .gen import (
    GenConf,
    GenRes,
    GenError,
    GenOut
)

from .thread import (
    Thread,
    Msg
)

from .json_schema import (
    JSchemaConf,
    json_schema_massage,
    json_schema_from_pydantic,
    pydantic_obj_from_json,
    build_dataclass_object_json_schema,
    build_root_json_schema,
    get_enum_type,
    get_final_type,
    create_final_instance
)

from .utils import (
    is_subclass_of,
    join_text
)




    
class Tokenizer(ABC):
    """Base tokenizer class to encode and decode tokens, measure text length in tokens, track special tokens."""

    bos_token_id: Union[int, None] # beginning of sentence
    bos_token: Union[str, None]

    eos_token_id: Union[int, None] # end of sentence
    eos_token: Union[str, None]

    pad_token_id: Union[int, None] # padding
    pad_token: Union[str, None]

    unk_token_id: Union[int, None] # unknown token
    unk_token: Union[str, None]

    
    def special_tokens(self) -> list[Any]:
        return [(self.bos_token_id, self.bos_token),
                (self.eos_token_id, self.eos_token),
                (self.pad_token_id, self.pad_token),
                (self.unk_token_id, self.unk_token)]
    
    def special_tokens_map(self) -> dict:
        return {"bos_token": self.bos_token,
                "eos_token": self.eos_token,
                "pad_token": self.pad_token,
                "unk_token": self.unk_token}
    
    @abstractmethod
    def encode(self, 
               text: str) -> list[int]:
        """Encode text into model tokens. Inverse of Decode().

        Args:
            text: Text to be encoded.

        Returns:
            A list of ints with the encoded tokens.
        """
        

    @abstractmethod
    def decode(self,
               token_ids: list[int],
               skip_special: bool = True) -> str:
        """Decode model tokens to text. Inverse of Encode().

        Args:
            token_ids: List of model tokens.
            skip_special: Don't decode special tokens like bos and eos. Defaults to True.

        Returns:
            Decoded text.
        """

    def token_len(self, 
                  text: str) -> int:
        """Returns token length for given text.

        Args:
            text: Text to be measured.

        Returns:
            Token length for given text.
        """

        tokens = self.encode(text)
        return len(tokens)        


    def __str__(self):
        return f"""\
bos={self.bos_token_id}='{self.bos_token}'
eos={self.eos_token_id}='{self.eos_token}'
pad={self.pad_token_id}='{self.pad_token}'
unk={self.unk_token_id}='{self.unk_token}'"""


    def __repr__(self):
        return str(self)







class Model(ABC):
    """Model is an abstract base class for common LLM model functionality. Many of the useful methods like extract() or json() are implemented here.

    It should not be instantiated directly, instead LlamaCppModel, OpenAIModel, etc, all derive from this class.
    """

    
    is_local_model: bool
    """Is the model running locally?"""
    
    is_message_model: bool
    """Is communication with the model message-based or text-based/token-based?"""

    tokenizer: Union[Tokenizer,None]
    """Tokenizer used to encode text. Some remote models don't have tokenizer and token length is estimated"""

    genconf: GenConf
    """Generation configuration: options used during gen()."""

    json_format_instructors: dict
    """If GenConf.json / GenConf.json_schema is used, these strings are appended to first thread msg of either instructions or IN kind. See initialization below."""

    json_in_dumps_kwargs: dict
    """Object to string formatting options for json.dumps() calls. See initialization below."""

    ctx_len: int
    """Maximum context token length, including input and model output. There can be a limit for output tokens in the max_tokens_limit."""

    max_tokens_limit: int
    """Some models limit the size of emitted output tokens."""

    maybe_image_input: bool
    """Does the model support images as input? A value of False is definitive, a value of True is actually a maybe, as some providers don't give this information. Check the model specs to be certain."""

    output_key_name: str
    """Name used when an output key needs to be created for JSON output."""

    output_fn_name: str
    """Function name for models that return JSON with a Tools/Functions-style API."""
    
    PROVIDER_NAME: str = NotImplemented
    """Provider prefix that this class handles."""

    
    def __init__(self,
                 is_local_model: bool,
                 genconf: Union[GenConf, None],
                 schemaconf: Union[JSchemaConf, None],
                 tokenizer: Union[Tokenizer, None]):
        """Initializer for base model type, shared by actual model classes like LlamaCpp, OpenAI, etc.

        Args:
            is_local_model: Is the model running locally?
            genconf: Default generation configuration options, used if generation call doesn't supply one.
            schemaconf: Default configuration for JSON schema validation, used if generation call doesn't supply one.
            tokenizer: Tokenizer used to encode text (even for message-based models).
        """
        
        self.is_local_model = is_local_model
        
        self.ctx_len = 0
        self.max_tokens_limit = sys.maxsize
        self.output_key_name = "output"
        self.output_fn_name = "json_out"

        self.tokenizer = tokenizer # type: ignore[assignment]

        if genconf is None:
            self.genconf = GenConf()
        else:
            self.genconf = genconf.clone()
        
        if schemaconf is None:
            self.schemaconf = JSchemaConf()
        else:
            self.schemaconf = schemaconf.clone()



        # set either "json" or "json_schema" key values to None to skip.
        self.json_format_instructors = {
            "json": {
                "bypass_if": ["json"], # bypass appending if all lowercase text values are present in thread
                "append_text": "Output JSON.",
                "sep_count": 2
            },
            "json_schema": {
                "bypass_if": ["json", "schema"],
                "append_text": "Output JSON matching the following schema:\n{{json_schema}}",
                "sep_count": 2
            }
        }

        # text going to model: tight, without \u00xx
        self.json_in_dumps_kwargs = {
            "indent": None,
            "ensure_ascii": False
        } 


    @abstractmethod
    def close(self):
        """Close model, release resources like memory or net connections."""
        ...




    
    # ======================================================== internal generation points    
            

    def gen(self,
            thread: Thread,
            genconf: Optional[GenConf] = None,
            ) -> GenOut:
        """Text generation from a Thread, used by the other model generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.
            NotImplementedError: If method was not defined by a derived class.

        Returns:
            A GenOut object with result, generated text, etc.
            The output text is in GenOut.text.
        """
        raise NotImplementedError


    async def gen_async(self,
                        thread: Thread,
                        genconf: Optional[GenConf] = None,
                        ) -> GenOut:
        """Async text generation from a Thread, used by the other model generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.
            NotImplementedError: If method was not defined by a derived class.

        Returns:
            A GenOut object with result, generated text, etc.
            The output text is in GenOut.text.
        """
        raise NotImplementedError
    


    
    
    def _gen_json_pre(self,
                      thread: Thread,
                      json_schema: Union[dict,str,None],
                      genconf: Union[GenConf,None],

                      massage_schema: bool,
                      schemaconf: Union[JSchemaConf, None]
                      ) -> list:

        if genconf is None:
            genconf = self.genconf

        if genconf.json_schema is not None and json_schema is not None:
            logger.warning("Both arg json_schema and genconf.json_schema are set: using json_schema arg")

        if json_schema is not None:
            if schemaconf is None:
                schemaconf = self.schemaconf

            logger.debug("JSON schema conf:\n" + pformat(schemaconf))

            if massage_schema:
                if not isinstance(json_schema, dict):
                    json_schema = json.loads(json_schema)

                json_schema = json_schema_massage(json_schema, schemaconf) # type: ignore[arg-type]
                logger.debug("Massaged JSON schema:\n" + pformat(json_schema))

        return [thread, genconf(format="json", 
                                json_schema=json_schema)]


    def gen_json(self,
                 thread: Thread,
                 json_schema: Union[dict,str,None],
                 genconf: Optional[GenConf] = None,

                 massage_schema: bool = True,
                 schemaconf: Optional[JSchemaConf] = None,
                 ) -> GenOut:
        """JSON/JSON-schema constrained generation, returning a Python dict of values, conditioned or not by a JSON schema.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread to use as model input.
            json_schema: A JSON schema describing the dict fields that will be output. None means no schema (free JSON output).
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            massage_schema: Simplify schema. Defaults to True.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with result, generated text, etc. The output dict is in GenOut.dic.
        """

        args = self._gen_json_pre(thread,
                                  json_schema,
                                  genconf,
                                  massage_schema,
                                  schemaconf)
        return self.gen(*args)
    

    async def gen_json_async(self,
                             thread: Thread,
                             json_schema: Union[dict,str,None],
                             genconf: Optional[GenConf] = None,

                             massage_schema: bool = True,
                             schemaconf: Optional[JSchemaConf] = None,
                             ) -> GenOut:
        """JSON/JSON-schema constrained generation, returning a Python dict of values, conditioned or not by a JSON schema.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread to use as model input.
            json_schema: A JSON schema describing the dict fields that will be output. None means no schema (free JSON output).
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            massage_schema: Simplify schema. Defaults to True.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with result, generated text, etc. The output dict is in GenOut.dic.
        """

        args = self._gen_json_pre(thread,
                                  json_schema,
                                  genconf,
                                  massage_schema,
                                  schemaconf)
        return await self.gen_async(*args)





    
    def _gen_dataclass_pre(self,
                           cls: Any # a dataclass
                           ) -> dict:
        
        if is_dataclass(cls):
            schema = build_dataclass_object_json_schema(cls)
        else:
            raise TypeError("Only dataclass allowed for cls argument")        
        return schema

    def _gen_dataclass_post(self,
                            out: GenOut,
                            cls: Any,
                            schemaconf: Union[JSchemaConf,None]
                            ) -> GenOut:
    
        if out.dic is not None:
            try:
                obj = create_final_instance(cls, 
                                            is_list=False,
                                            val=out.dic,
                                            schemaconf=schemaconf)
                out.value = obj
                
            except TypeError as e:
                out.res = GenRes.ERROR_JSON_SCHEMA_VAL # error initializing object from JSON
                out.text += f"\nJSON Schema error: {e}"
        else:
            # out.res already holds the right error
            ...
        
        return out




    def gen_dataclass(self,
                      cls: Any, # a dataclass
                      thread: Thread,
                      genconf: Optional[GenConf] = None,
                      schemaconf: Optional[JSchemaConf] = None
                      ) -> GenOut:
        """Constrained generation after a dataclass definition.
        An initialized dataclass object is returned in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut containing the created object.

        Args:
            cls: A dataclass definition.
            thread: The Thread object to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with result, generated text, etc. The initialized dataclass object is in GenOut.value.
        """

        schema = self._gen_dataclass_pre(cls)

        out = self.gen_json(thread,
                            schema,
                            genconf,
                            massage_schema=True,
                            schemaconf=schemaconf)
    
        return self._gen_dataclass_post(out,
                                        cls,
                                        schemaconf)



    async def gen_dataclass_async(self,
                                  cls: Any, # a dataclass
                                  thread: Thread,
                                  genconf: Optional[GenConf] = None,
                                  schemaconf: Optional[JSchemaConf] = None
                                  ) -> GenOut:
        """Async constrained generation after a dataclass definition.
        An initialized dataclass object is returned in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut containing the created object.

        Args:
            cls: A dataclass definition.
            thread: The Thread object to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with result, generated text, etc. The initialized dataclass object is in GenOut.value.
        """

        schema = self._gen_dataclass_pre(cls)

        out = await self.gen_json_async(thread,
                                        schema,
                                        genconf,
                                        massage_schema=True,
                                        schemaconf=schemaconf)
    
        return self._gen_dataclass_post(out,
                                        cls,
                                        schemaconf)







    def _gen_pydantic_pre(self,
                          cls: Any # a Pydantic BaseModel class
                          ) -> dict:

        if is_subclass_of(cls, BaseModel):
            schema = json_schema_from_pydantic(cls)
        else:
            raise TypeError("Only pydantic BaseModel allowed for cls argument")
        
        return schema


    def _gen_pydantic_post(self,
                           out: GenOut,
                           cls: Any, # a Pydantic BaseModel class
                           schemaconf: Union[JSchemaConf,None]
                           ) -> GenOut:
    
        if out.dic is not None:
            try:
                obj = pydantic_obj_from_json(cls, 
                                             out.dic,
                                             schemaconf=schemaconf)
                out.value = obj
                
            except TypeError as e:
                out.res = GenRes.ERROR_JSON_SCHEMA_VAL # error validating for object (by Pydantic), but JSON is valid for its schema
                out.text += f"\nJSON Schema error: {e}"
        else:
            # out.res already holds the right error
            ...
        
        return out




    def gen_pydantic(self,
                     cls: Any, # a Pydantic BaseModel class
                     thread: Thread,
                     genconf: Optional[GenConf] = None,
                     schemaconf: Optional[JSchemaConf] = None
                     ) -> GenOut:
        """Constrained generation after a Pydantic BaseModel-derived class definition.
        An initialized Pydantic BaseModel object is returned in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut containing the created object.

        Args:
            cls: A class derived from a Pydantic BaseModel class.
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.
            TypeError: When cls is not a Pydantic BaseClass.

        Returns:
            A GenOut object with result, generated text, etc. The initialized Pydantic BaseModel-derived object is in GenOut.value.
        """

        schema = self._gen_pydantic_pre(cls)

        out = self.gen_json(thread,
                            schema,
                            genconf,
                            massage_schema=True,
                            schemaconf=schemaconf)
    
        return self._gen_pydantic_post(out,
                                       cls,
                                       schemaconf)



    async def gen_pydantic_async(self,
                                 cls: Any, # a Pydantic BaseModel class
                                 thread: Thread,
                                 genconf: Optional[GenConf] = None,
                                 schemaconf: Optional[JSchemaConf] = None
                                 ) -> GenOut:
        """Async constrained generation after a Pydantic BaseModel-derived class definition.
        An initialized Pydantic BaseModel object is returned in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut containing the created object.

        Args:
            cls: A class derived from a Pydantic BaseModel class.
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.
            TypeError: When cls is not a Pydantic BaseClass.

        Returns:
            A GenOut object with result, generated text, etc. The initialized Pydantic BaseModel-derived object is in GenOut.value.
        """

        schema = self._gen_pydantic_pre(cls)

        out = await self.gen_json_async(thread,
                                        schema,
                                        genconf,
                                        massage_schema=True,
                                        schemaconf=schemaconf)

        return self._gen_pydantic_post(out,
                                       cls,
                                       schemaconf)






    def _gen_extract_pre(self,
                         target: Any,
                         thread: Thread,
                         genconf: Union[GenConf, None],
                         schemaconf: Union[JSchemaConf, None]
                         ) -> Any:

        schema, created_output_key = build_root_json_schema(target, 
                                                            self.output_key_name)
        final_type, is_list = get_final_type(target)

        if schemaconf is None:
            schemaconf = JSchemaConf()

        return (thread,
                schema,
                genconf,                
                created_output_key,
                final_type, 
                is_list,
                schemaconf)


    def _gen_extract_post(self,
                          out: GenOut,
                          created_output_key: bool,
                          final_type: Any, 
                          is_list: bool,
                          schemaconf: JSchemaConf
                          ) -> GenOut:

        if out.dic is not None:

            if created_output_key:
                if self.output_key_name in out.dic:
                    val = out.dic[self.output_key_name]
                else:
                    out.res = GenRes.ERROR_JSON_SCHEMA_VAL # JSON error
                    out.text += f"\nExpecting key '{self.output_key_name}'."
                    return out

            else:
                val = out.dic

            try:
                value = create_final_instance(final_type, 
                                              is_list, 
                                              val,
                                              schemaconf=schemaconf)
                out.value = value

            except TypeError as e:
                out.res = GenRes.ERROR_JSON_SCHEMA_VAL # error validating, but JSON is valid for its schema
                out.text += f"\nJSON Schema error: {e}"

        else:
            # out.res already holds the right error
            ...

        return out



    def gen_extract(self,
                    target: Any,
                    thread: Thread,
                    genconf: Optional[GenConf] = None,
                    schemaconf: Optional[JSchemaConf] = None
                    ) -> GenOut:
        """Free type constrained generation: an instance of the given type is initialized with the model's output.
        The initialized value is placed in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        The following target types are accepted:

            - prim_type:
                bool
                int
                float
                str
                
            - enums:
                [1, 2, 3] or ["a","b"] - all items of the same prim_type
                Literal['year', 'name'] - all items of the same prim_type
                Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

            - datetime/date/time

            - a list in the form:
                list[type] - for example list[int]. 
                The list can be annotated:
                    Annotated[list[T], "List desc"]
                And/or the list item type can be annotated:
                    list[Annotated[T, "Item desc"]]

            - dataclass with fields of the above supported types (or dataclass).

            - Pydantic BaseModel

        All types can be Annotated[T, "Desc"], for example: 
            count: int
        Can be annotated as:
            count: Annotated[int, "How many units?"]

        Args:
            target: One of the above types.
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with model's results and instantiated type in the "value" field.
        """

        (thread,
         schema,
         genconf, 
         created_output_key, 
         final_type, 
         is_list, 
         schemaconf) = self._gen_extract_pre(target,
                                             thread,
                                             genconf,
                                             schemaconf)

        out = self.gen_json(thread,
                            schema,
                            genconf,
                            massage_schema=True,
                            schemaconf=schemaconf)
    
        return self._gen_extract_post(out,
                                      created_output_key,
                                      final_type,
                                      is_list,
                                      schemaconf) # type: ignore [arg-type]




    async def gen_extract_async(self,
                                target: Any,
                                thread: Thread,
                                genconf: Optional[GenConf] = None,
                                schemaconf: Optional[JSchemaConf] = None
                                ) -> GenOut:
        """Async free type constrained generation: an instance of the given type is initialized with the model's output.
        The initialized value is placed in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        The following target types are accepted:

            - prim_type:
                bool
                int
                float
                str
                
            - enums:
                [1, 2, 3] or ["a","b"] - all items of the same prim_type
                Literal['year', 'name'] - all items of the same prim_type
                Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

            - datetime/date/time

            - a list in the form:
                list[type] - for example list[int]. 
                The list can be annotated:
                    Annotated[list[T], "List desc"]
                And/or the list item type can be annotated:
                    list[Annotated[T, "Item desc"]]

            - dataclass with fields of the above supported types (or dataclass).

            - Pydantic BaseModel

        All types can be Annotated[T, "Desc"], for example: 
            count: int
        Can be annotated as:
            count: Annotated[int, "How many units?"]

        Args:
            target: One of the above types.
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with model's results and instantiated type in the "value" field.
        """
        (thread,
         schema,
         genconf, 
         created_output_key, 
         final_type, 
         is_list, 
         schemaconf) = self._gen_extract_pre(target,
                                             thread,
                                             genconf,
                                             schemaconf)

        out = await self.gen_json_async(thread,
                                        schema,
                                        genconf,
                                        massage_schema=True,
                                        schemaconf=schemaconf)
    
        return self._gen_extract_post(out,
                                      created_output_key,
                                      final_type,
                                      is_list,
                                      schemaconf) # type: ignore [arg-type]







    # ======================================================== user generation points    


    def __call__(self,             
                 query: Union[Thread,Msg,tuple,str],
                 *,
                 inst: Optional[str] = None,

                 genconf: Optional[GenConf] = None,
                 ok_length_is_error: bool = False
                 ) -> str:
        """Text generation from a Thread or plain text, used by the other model generation methods. Same as call().

        Args:
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            ok_length_is_error: Should a result of GenRes.OK_LENGTH be considered an error and raise?

        Raises:
            GenError: If an error occurred. This can be a model error, or an invalid JSON output error.
            RuntimeError: If unable to generate.

        Returns:
            Text generated by model.
        """

        return self.call(query,
                         inst=inst,
                         genconf=genconf,
                         ok_length_is_error=ok_length_is_error)


    def call(self,             
             query: Union[Thread,Msg,tuple,str],
             *,
             inst: Optional[str] = None,

             genconf: Optional[GenConf] = None,
             ok_length_is_error: bool = False
             ) -> str:
        """Text generation from a Thread or plain text, used by the other model generation methods.

        Args:
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            ok_length_is_error: Should a result of GenRes.OK_LENGTH be considered an error and raise?

        Raises:
            GenError: If an error occurred. This can be a model error, or an invalid JSON output error.
            RuntimeError: If unable to generate.

        Returns:
            Text generated by model.
        """
        
        thread = Thread.ensure(query, inst)

        out = self.gen(thread=thread, 
                       genconf=genconf)
        
        GenError.raise_if_error(out,
                                ok_length_is_error=ok_length_is_error)

        return out.text


    async def call_async(self,
                         query: Union[Thread,Msg,tuple,str],
                         *,
                         inst: Optional[str] = None,
 
                         genconf: Optional[GenConf] = None,
                         ok_length_is_error: bool = False
                         ) -> str:
        """Text generation from a Thread or plain text, used by the other model generation methods.

        Args:
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            ok_length_is_error: Should a result of GenRes.OK_LENGTH be considered an error and raise?

        Raises:
            GenError: If an error occurred. This can be a model error, or an invalid JSON output error.
            RuntimeError: If unable to generate.

        Returns:
            Text generated by model.
        """
        
        thread = Thread.ensure(query, inst)

        out = await self.gen_async(thread=thread, 
                                   genconf=genconf)
        
        GenError.raise_if_error(out,
                                ok_length_is_error=ok_length_is_error)

        return out.text









    def json(self,
             query: Union[Thread,Msg,tuple,str],
             *,
             json_schema: Union[dict,str,None] = None,
             inst: Optional[str] = None,

             genconf: Optional[GenConf] = None,
             massage_schema: bool = True,
             schemaconf: Optional[JSchemaConf] = None,
             ) -> dict:
        """JSON/JSON-schema constrained generation, returning a Python dict of values, constrained or not by a JSON schema.
        Raises GenError if unable to get a valid/schema-validated JSON.

        Args:
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            json_schema: A JSON schema describing the dict fields that will be output. None means no schema (free JSON output).
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            massage_schema: Simplify schema. Defaults to True.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example an invalid JSON schema output error. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            A dict from model's JSON response, following genconf.jsonschema, if provided.
        """        

        thread = Thread.ensure(query, inst)

        out = self.gen_json(thread,
                            json_schema,                            
                            genconf,
                            massage_schema,
                            schemaconf)
        
        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.dic # type: ignore[return-value]






    async def json_async(self,             
                         query: Union[Thread,Msg,tuple,str],
                         *,
                         json_schema: Union[dict,str,None] = None,
                         inst: Optional[str] = None,

                         genconf: Optional[GenConf] = None,
                         massage_schema: bool = True,
                         schemaconf: Optional[JSchemaConf] = None,
                         ) -> dict:
        """JSON/JSON-schema constrained generation, returning a Python dict of values, constrained or not by a JSON schema.
        Raises GenError if unable to get a valid/schema-validated JSON.

        Args:
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            json_schema: A JSON schema describing the dict fields that will be output. None means no schema (free JSON output).
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            massage_schema: Simplify schema. Defaults to True.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example an invalid JSON schema output error. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            A dict from model's JSON response, following genconf.jsonschema, if provided.
        """        

        thread = Thread.ensure(query, inst)

        out = await self.gen_json_async(thread,
                                        json_schema,
                                        genconf,
                                        massage_schema,
                                        schemaconf)
        
        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.dic # type: ignore[return-value]











    def dataclass(self, # noqa: F811
                  cls: Any, # a dataclass definition

                  query: Union[Thread,Msg,tuple,str],
                  *,
                  inst: Optional[str] = None,

                  genconf: Optional[GenConf] = None,
                  schemaconf: Optional[JSchemaConf] = None
                  ) -> Any: # a dataclass object
        """Constrained generation after a dataclass definition, resulting in an object initialized with the model's response.
        Raises GenError if unable to get a valid response that follows the dataclass definition.

        Args:
            cls: A dataclass definition.
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example invalid object initialization. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            An object of class cls (derived from dataclass) initialized from the constrained JSON output.
        """

        thread = Thread.ensure(query, inst)

        out = self.gen_dataclass(cls,
                                 thread,
                                 genconf,
                                 schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value



    async def dataclass_async(self, # noqa: E811
                              cls: Any, # a dataclass definition

                              query: Union[Thread,Msg,tuple,str],
                              *,
                              inst: Optional[str] = None,

                              genconf: Optional[GenConf] = None,
                              schemaconf: Optional[JSchemaConf] = None
                              ) -> Any: # a dataclass object
        """Async constrained generation after a dataclass definition, resulting in an object initialized with the model's response.
        Raises GenError if unable to get a valid response that follows the dataclass definition.

        Args:
            cls: A dataclass definition.
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example invalid object initialization. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            An object of class cls (derived from dataclass) initialized from the constrained JSON output.
        """

        thread = Thread.ensure(query, inst)

        out = await self.gen_dataclass_async(cls,
                                             thread,
                                             genconf,
                                             schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value









    def pydantic(self,
                 cls: Any, # a Pydantic BaseModel class

                 query: Union[Thread,Msg,tuple,str],
                 *,
                 inst: Optional[str] = None,

                 genconf: Optional[GenConf] = None,
                 schemaconf: Optional[JSchemaConf] = None
                 ) -> Any: # a Pydantic BaseModel object
        """Constrained generation after a Pydantic BaseModel-derived class definition.
        Results in an object initialized with the model response.
        Raises GenError if unable to get a valid dict that follows the BaseModel class definition.

        Args:
            cls: A class derived from a Pydantic BaseModel class.
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example an invalid BaseModel object. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            A Pydantic object of class cls (derived from BaseModel) initialized from the constrained JSON output.
        """

        thread = Thread.ensure(query, inst)

        out = self.gen_pydantic(cls,
                                thread,
                                genconf,
                                schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value


    async def pydantic_async(self,
                             cls: Any, # a Pydantic BaseModel class

                             query: Union[Thread,Msg,tuple,str],
                             *,
                             inst: Optional[str] = None,

                             genconf: Optional[GenConf] = None,
                             schemaconf: Optional[JSchemaConf] = None
                             ) -> Any: # a Pydantic BaseModel object
        """Async constrained generation after a Pydantic BaseModel-derived class definition.
        Results in an object initialized with the model response.
        Raises GenError if unable to get a valid dict that follows the BaseModel class definition.

        Args:
            cls: A class derived from a Pydantic BaseModel class.
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example an invalid BaseModel object. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            A Pydantic object of class cls (derived from BaseModel) initialized from the constrained JSON output.
        """

        thread = Thread.ensure(query, inst)

        out = await self.gen_pydantic_async(cls,
                                            thread,
                                            genconf,
                                            schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value







    def extract(self,
                target: Any,

                query: Union[Thread,Msg,tuple,str],
                *,
                inst: Optional[str] = None,

                genconf: Optional[GenConf] = None,
                schemaconf: Optional[JSchemaConf] = None
                ) -> Any:        
        """Type-constrained generation: an instance of the given type will be initialized with the model's output.
        The following target types are accepted:

        - prim_type:

            - bool
            - int
            - float
            - str
            
        - enums:

            - [1, 2, 3] or ["a","b"] - all items of the same prim_type
            - Literal['year', 'name'] - all items of the same prim_type
            - Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

        - datetime/date/time

        - a list in the form:
            - list[type]
            
            For example list[int]. The list can be annotated:
                Annotated[list[T], "List desc"]
            And/or the list item type can be annotated:
                list[Annotated[T, "Item desc"]]

        - dataclass with fields of the above supported types (or dataclass).

        - Pydantic BaseModel

        All types can be Annotated[T, "Desc"], for example: 
            count: int
        Can be annotated as:
            count: Annotated[int, "How many units?"]

        Args:
            target: One of the above types.
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example invalid object initialization. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            A value of target arg type instantiated with the model's output.
        """

        thread = Thread.ensure(query, inst)

        out = self.gen_extract(target,
                               thread,
                               genconf,
                               schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value





    async def extract_async(self,
                            target: Any,

                            query: Union[Thread,Msg,tuple,str],
                            *,
                            inst: Optional[str] = None,

                            genconf: Optional[GenConf] = None,
                            schemaconf: Optional[JSchemaConf] = None
                            ) -> Any:        
        """Async type-constrained generation: an instance of the given type will be initialized with the model's output.
        The following target types are accepted:

        - prim_type:

            - bool
            - int
            - float
            - str
            
        - enums:

            - [1, 2, 3] or ["a","b"] - all items of the same prim_type
            - Literal['year', 'name'] - all items of the same prim_type
            - Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

        - datetime/date/time

        - a list in the form:
            - list[type]
            
            For example list[int]. The list can be annotated:
                Annotated[list[T], "List desc"]
            And/or the list item type can be annotated:
                list[Annotated[T, "Item desc"]]

        - dataclass with fields of the above supported types (or dataclass).

        - Pydantic BaseModel

        All types can be Annotated[T, "Desc"], for example: 
            count: int
        Can be annotated as:
            count: Annotated[int, "How many units?"]

        Args:
            target: One of the above types.
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example invalid object initialization. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            A value of target arg type instantiated with the model's output.
        """

        thread = Thread.ensure(query, inst)

        out = await self.gen_extract_async(target,
                                           thread,
                                           genconf,
                                           schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value









    def classify(self,
                 labels: Any,

                 query: Union[Thread,Msg,tuple,str],
                 *,
                 inst: Optional[str] = None,

                 genconf: Optional[GenConf] = None,
                 schemaconf: Optional[JSchemaConf] = None
                 ) -> Any:
        """Returns a classification from one of the given enumeration values
        The following ways to specify the valid labels are accepted:

        - [1, 2, 3] or ["a","b"] - all items of the same prim_type
        - Literal['year', 'name'] - all items of the same prim_type
        - Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

        Args:
            labels: One of the above types.
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            One of the given labels, as classified by the model.
        """

        # verify it's a valid enum "type"
        type_,_ = get_enum_type(labels)
        if type_ is None:
            raise TypeError("Arg labels must be one of Literal, Enum class or a list of str, float or int items")
        
        return self.extract(labels,
                            query,
                            inst=inst,
                            genconf=genconf,
                            schemaconf=schemaconf)



    async def classify_async(self,
                             labels: Any,

                             query: Union[Thread,Msg,tuple,str],
                             *,
                             inst: Optional[str] = None,

                             genconf: Optional[GenConf] = None,
                             schemaconf: Optional[JSchemaConf] = None
                             ) -> Any:
        """Returns a classification from one of the given enumeration values
        The following ways to specify the valid labels are accepted:

        - [1, 2, 3] or ["a","b"] - all items of the same prim_type
        - Literal['year', 'name'] - all items of the same prim_type
        - Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

        Args:
            labels: One of the above types.
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred. See GenError.
            RuntimeError: If unable to generate.

        Returns:
            One of the given labels, as classified by the model.
        """

        # verify it's a valid enum "type"
        type_,_ = get_enum_type(labels)
        if type_ is None:
            raise TypeError("Arg labels must be one of Literal, Enum class or a list of str, float or int items")
        
        return await self.extract_async(labels,
                                        query,
                                        inst=inst,
                                        genconf=genconf,
                                        schemaconf=schemaconf)









    
    # ======================================================== properties

     

    @abstractmethod
    def token_len(self,
                  thread_or_text: Union[Thread,str],
                  genconf: Optional[GenConf] = None) -> int:
        """Calculate or estimate the token length for a Thread or a plain text string.
        In some cases where it's not possible to calculate the exact token count, 
        this function should give a conservative (upper bound) estimate.
        It's up to the implementation whether to account for side information like JSON Schema,
        but it must reflect the model's context token accounting.
        Thread or text must be the final text which will passed to model.

        Args:
            thread_or_text: Final thread or text to be passed to model.
            genconf: Model generation configuration. Defaults to None.

        Returns:
            Number of tokens occupied.
        """
        ...



    @property
    def token_len_lambda(self) -> Callable[[Union[Thread,str],Optional[GenConf]], int]:
        return lambda thread_or_text, genconf=None: self.token_len(thread_or_text, genconf) # type: ignore[misc]




    def calc_max_max_tokens(self,
                            input_token_len: int) -> int:
        """Calculate maximum length of tokens that the model can output (as in GenConf.max_tokens).
        May be smaller than remaining length due to max_tokens_limit attribute.

        Args:
            input_token_len: Length in tokens of model input.

        Returns:
            Maximum tokens that can be generated by model.
        """

        token_len = self.ctx_len - input_token_len
        return min(token_len, self.max_tokens_limit)



    def resolve_genconf_max_tokens(self,
                                   input_token_len: int,
                                   genconf: GenConf) -> int:
        """Resolve genconf.max_tokens to a definitive value, depending on input, ctx_len and max_tokens_limit"""
        
        avail_output_tokens = self.calc_max_max_tokens(input_token_len)
        if avail_output_tokens <= 0:
            raise ValueError(f"""Input token length ({input_token_len}) doesn't fit available ctx_len ({self.ctx_len}) or max_tokens_limit ({self.max_tokens_limit})""")

        # calc maximum possible output
        resolved_max_tokens = genconf.resolve_max_tokens(self.ctx_len, self.max_tokens_limit)

        # ensure avail_output_tokens fits resolved_max_tokens
        max_tokens = min(resolved_max_tokens, avail_output_tokens)
        return max_tokens



    @classmethod
    def known_models(cls,
                     api_key: Optional[str] = None) -> Union[list[str], None]:
        """If the model can only use a fixed set of models, return their names. Otherwise, return None.

        Args:
            api_key: If the model provider requires an API key, pass it here or set it in the respective env variable.

        Returns:
            Returns a list of known models or None if unable to fetch it.
        """
        return None





    @abstractmethod
    def name(self) -> str:
        """Model (short) name."""
        ...

    @abstractmethod
    def desc(self) -> str:
        """Model description."""
        ...
    
    def info(self) -> str:
        """Model description and config information."""
        return f"desc='{self.desc()}',\n" \
               f"ctx_len={self.ctx_len}, max_tokens_limit={self.max_tokens_limit},\n" \
               f"genconf={self.genconf}"
               

    @classmethod
    @abstractmethod
    def provider_version(cls) -> str:
        """Provider library version: provider x.y.z
        Ex. llama-cpp-python-0.2.44
        """
        ...

    @classmethod
    def version(cls) -> str:
        """Sibila version + provider version
        Ex: sibila=0.2.3 provider=llama-cpp-python-0.2.44
        """        
        from .__init__ import __version__ as version  # type: ignore[import-not-found]
        return f"sibila={version} provider={cls.provider_version()}"
        


    def __str__(self):
        return self.desc_info()





    
    
    # = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = internal
    
    def _prepare_gen_thread(self,
                            thread: Thread,
                            genconf: GenConf) -> Thread:
        """Perform common pre-generation operations of the thread that will be used.

        Args:
            thread: Input thread that may be augmented here. If not modified, the same thread is returned.
            genconf: Generation configuration.

        Raises:
            ValueError: If empty thread.

        Returns:
            A thread with A GenOut object with result, generated text, etc.
        """

        logger.debug(f"GenConf: {genconf}")

        if not len(thread):
            raise ValueError("Cannot generate from an empty thread")
        
        if thread[0].kind != Msg.Kind.IN:
            logger.warning("First thread message is not an IN message.")
        if thread[-1].kind != Msg.Kind.IN:
            logger.warning("Last thread message is not an IN message.")

        if genconf.format == "json":
            if genconf.json_schema is None:
                inst = self.json_format_instructors["json"]
            else:
                inst = self.json_format_instructors["json_schema"]

            if inst is not None:
                bypass = all([thread.has_text_lower(t) for t in inst["bypass_if"]])
    
                if not bypass:
                    thread = thread.clone()
                    text = inst["append_text"]
                    
                    if genconf.json_schema is not None:
                        if isinstance(genconf.json_schema, str): 
                            sc = genconf.json_schema
                        else:
                            sc = json.dumps(genconf.json_schema, **self.json_in_dumps_kwargs)

                        text = text.replace("{{json_schema}}", sc)

                    first_text = thread[0].text
                    first_text = join_text(first_text, 
                                           text,
                                           thread.join_sep * inst["sep_count"]) # will separate by join_sep * sep_count
                    thread[0].text = first_text
                    
                    logger.debug(f'Appended {"json" if genconf.json_schema is None else "json_schema"} instruction to first message.')

            
        logger.debug(f"Prepare: {str(thread)}")
        
        return thread


    
    
    def _prepare_gen_out(self,
                         response: Union[str,dict, None],
                         finish: str,
                         genconf: GenConf) -> GenOut:
        """Perform common operations over the generated model response.

        Args:
            response: Text or a JSON object obtained from model.
            finish: Model finish reason - only "stop", "length" are used, others will map as GenRes.ERROR_MODEL.

        Returns:
            A GenOut object with result, generated text, etc. 
        """

        logger.debug(f"Response finish='{finish}': █{response}█")
        
        if response is None:
            response = ''

        if genconf.format == "json":
            if isinstance(response, str):
                response = response.strip()

                if "\\u" in response:
                    # dumps(with default ensure_ascii=False) -> ascii (subset of utf-8) -> text.encode("latin1") -> latin1 ->
                    #   decode("unicode-escape") -> utf-8
                    response = response.encode("latin1").decode("unicode-escape")

                # some troubled remote models may include chit-chat after the JSON
                begin = response.find("{")
                if begin > 0:
                    response = response[begin:]
                end = response.rfind("}")
                if end > 0:
                    response = response[:end + 1]

            try:
                if isinstance(response, str):
                    dic = json.loads(response)
                else:
                    dic = response

                if genconf.json_schema is not None:
                    if isinstance(genconf.json_schema, str):
                        schema_dic = json.loads(genconf.json_schema)
                    else:
                        schema_dic = genconf.json_schema
                        
                    json_schema_validate(dic, schema_dic)

                out = GenOut(res=GenRes.from_finish_reason(finish),
                             text=str(response),
                             dic=dic)
            
            except json.JSONDecodeError as err:
                out = GenOut(res=GenRes.from_finish_reason("!json"),
                             text=f"'{err}' {response}")
                
            except json_schema_ValidationError as err:
                logger.info(f"JSON schema validation error: {err.message}")
                out = GenOut(res=GenRes.from_finish_reason("!json_schema_val"),
                             text=f"'{err.message}' {response}",
                             dic=dic)
                
            except json_schema_SchemaError as err:
                logger.info(f"JSON schema error: {err.message}")
                out = GenOut(res=GenRes.from_finish_reason("!json_schema_error"),
                             text=f"'{err.message}' {response}",
                             dic=dic)

        else:
            if isinstance(response, str):
                response = response.strip()
            else:
                response = str(response)

            out = GenOut(res=GenRes.from_finish_reason(finish),
                         text=response)
            
       
        return out
    








from .models import Models


class FormattedTextModel(Model, ABC):
    """Model that uses formatted text (chat templates) for input/output."""
    
    format_template: Union[str,None]
    _jinja_compiled_template: Union[Any,None]

    def __init__(self,
                 is_local_model: bool,
                 genconf: Union[GenConf, None],
                 schemaconf: Union[JSchemaConf, None],
                 tokenizer: Union[Tokenizer, None]):

        super().__init__(is_local_model,
                         genconf,
                         schemaconf,
                         tokenizer)

        self.is_message_model = False

        self.format_template = None
        self._jinja_compiled_template = None
        
    
    def init_format(self,
                    format: Union[str,None],
                    format_search_order: list,
                    info: dict):
        """Search for a chat template format suitable for model.

        Arg format_search_order is a list of str flags:
            "name: match by name
            "meta_template": use model file metadata's template
            "folder_json": search from "models.json" and/or "formats.json" in the same folder
            ex: ["name", "meta_template", "folder_json"]
        
        Arg info required fields: {
            "name": "filename.gguf"
            "path": full path to model
            "meta_template_name": "chat_template"
        }

        Args:
            format: Format name or actual Jinja template.
            format_search_order: List of flags specifying an order for format searching.
            info: Extra data for format searching.

        Raises:
            ValueError: If a chat template format was not found.

        Returns:
            A Jinja template format.
        """

        # handle chat template format
        def search_format() -> str:

            for order in format_search_order:

                if order == "name":
                    fmt = Models.match_format_template(info["name"])
                    if fmt is not None:
                        return fmt
                        
                elif order == "meta_template":
                    md = self.get_metadata()
                    key_name = info["meta_template_name"]
                    if key_name in md:
                        logger.debug(f"Format from model file metadata ('{key_name}') template='{md[key_name]}'")
                        fmt = md[key_name]
                        return fmt # type: ignore[return-value]
                    
                elif order == "folder_json":
                    fmt = Models.folder_match_format_template(info["path"])
                    if fmt is not None:
                        return fmt

            raise ValueError(f"Could not find chat template format for model '{info['name']}'. "
                             "Fine-tuned models cannot work well without the right format. "
                             "Please provide a chat template in the 'format' argument: either as a Jinja template, or the format name if already defined in Models' formats.json. "
                             "See the docs for more information.")

        
        if format is not None: # format was passed (call or Models' model entry)
            if '{{' in format: # an str with a jinja template
                self.format_template = format
                
            else: # a format name
                self.format_template = Models.get_format_template(format)
                if self.format_template is None:
                    logger.warning(f"Could not find format name '{format}'. Will further search in model's metadata and own folder.")
                
        if self.format_template is None: # attempt to find format by other means
            self.format_template = search_format() # will raise if unable to find


        # setup jinja template
        jinja_env = ImmutableSandboxedEnvironment(trim_blocks=True, lstrip_blocks=True)
        jinja_env.globals["raise_exception"] = jinja_raise_exception
        self._jinja_compiled_template = jinja_env.from_string(self.format_template)






    def gen(self, 
            thread: Thread,
            genconf: Optional[GenConf] = None,
            ) -> GenOut:
        """Text generation from a Thread, used by the other model generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread object to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.

        Raises:
            ValueError: If trying to generate from an empty prompt.
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with result, generated text, etc. 
        """

        if genconf is None:
            genconf = self.genconf

        text,finish = self._gen_thread(thread, genconf)

        return self._prepare_gen_out(text, finish, genconf)




    async def gen_async(self, 
                        thread: Thread,
                        genconf: Optional[GenConf] = None,
                        ) -> GenOut:
        """Asynchronous generation from a Thread, used by the other model generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread object to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.

        Raises:
            ValueError: If trying to generate from an empty prompt.
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with result, generated text, etc. 
        """

        if genconf is None:
            genconf = self.genconf

        text,finish = await self._gen_thread_async(thread, genconf)

        return self._prepare_gen_out(text, finish, genconf)



    
    @abstractmethod
    def _gen_thread(self,
                    thread: Thread,
                    genconf: GenConf) -> tuple[str,str]:
        """Generate from formatted text.

        Args:
            thread: Input Thread object.
            genconf: Model generation configuration.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            Tuple of strings: generated_text, finish_reason.
        """
        ...


    @abstractmethod
    async def _gen_thread_async(self,
                                thread: Thread,
                                genconf: GenConf) -> tuple[str,str]:
        """Asynchronously generate from formatted text.

        Args:
            thread: Input Thread object.
            genconf: Model generation configuration.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            Tuple of strings: generated_text, finish_reason.
        """
        ...



    def get_metadata(self) -> dict:
        """Returns model metadata."""
        return {}













class MessagesModel(Model, ABC):
    """Model with message-based communication."""

    def __init__(self,
                 is_local_model: bool,
                 genconf: Union[GenConf,None],
                 schemaconf: Union[JSchemaConf,None],
                 tokenizer: Union[Tokenizer, None]):

        super().__init__(is_local_model,
                         genconf,
                         schemaconf,
                         tokenizer)
        
        self.is_message_model = True
                


    def resolve_settings(self,
                         provider: str,
                         name: str,
                         keys: list[str]) -> dict:
        """
        Return: dict with values for found keys.            
        """

        # 1: locate a provider:name entry in Models (in res/base_models.json or overridden by user settings)
        res_name = provider + ":" + name
        provider, name, args = Models.resolve_model_entry(res_name)

        logger.debug(f"Resolved '{res_name}' to '{provider}:{name}' with defaults: {args}")

        # 2: update located defaults with args
        out = {}
        for key in keys:
            if key in args:
                out[key] = args[key]
        return out






"""
Requires a defined env variable ANTHROPIC_API_KEY with a valid key.
"""

import pytest

import os, subprocess, shutil, json, asyncio
from typing import Any, Optional, Union, Literal, Annotated, get_origin, get_args

import logging
logger = logging.getLogger(__name__)
# pytest --log-cli-level=DEBUG

from dotenv import load_dotenv


from sibila import (
    AnthropicModel,
    Models,
    GenConf,
    Thread,
    Msg
)



from .utils import setup_env_models, teardown_env_models, setup_model, teardown_model
from .utils import run_cmd, run_text, run_json


MODEL_NAME = "claude-3-haiku-20240307"


DO_TEARDOWN = True

@pytest.fixture(autouse=True, scope="module")
def env_model():

    load_dotenv(override=True, verbose=True)

    base_dir, models_dir = setup_env_models("anthropic-" + MODEL_NAME, 
                                            change_cwd=True,
                                            full_clean=False)

    ret = base_dir, models_dir
    print("---> setup", ret)
    
    yield ret

    # --------------------------- teardown
    if DO_TEARDOWN == False:
        return

    print("---> teardown", ret)

    teardown_env_models(base_dir)







describe_ops = [
    ("../res/coypu.jpg", "boat"),
    ("https://upload.wikimedia.org/wikipedia/commons/thumb/3/3b/Singapura_Cats.jpg/320px-Singapura_Cats.jpg", "cats")
]

def test_describe(env_model):

    model = AnthropicModel(MODEL_NAME)

    for op in describe_ops:
        th = Thread(Msg.make_IN("Describe this image", op[0]))
        res = model(th)
        print(res)
        assert op[1] in res.lower()


count_ops = [
    ("How many elements in this image?", "https://upload.wikimedia.org/wikipedia/commons/thumb/3/3b/Singapura_Cats.jpg/320px-Singapura_Cats.jpg", 2),
    ("How many persons in this image?", "https://upload.wikimedia.org/wikipedia/en/6/6e/VUMillennium.jpg", 4)
]

def test_count(env_model):

    model = AnthropicModel(MODEL_NAME)

    for op in count_ops:
        th = Thread(Msg.make_IN(op[0], op[1]))
        res = model.extract(int, th)
        print(res)
        assert res == op[2]






import pytest

from typing import Any, Optional, Union, Literal, Annotated, get_origin, get_args
import enum

from datetime import date, time, datetime

from dataclasses import dataclass, field

from pydantic import BaseModel, Field

from sibila.json_schema import (
    JSchemaConf,
    build_root_json_schema,
    json_schema_massage,
)



def test_json_schema_pydantic():

    class Person(BaseModel):
        name: str  
        occupation: str = Field(description="A peaceful occupation")

        # Optional, Union
        age0: Optional[int]
        age1: Optional[int] = Field(description="Description for age1")

        option0: Union[int,str]
        option1: Union[int,str] = Field(description="Description for option1")

        # defaults start here
        age2: Optional[int] = Field(default=None)
        age3: Optional[int] = Field(description="Description for age3", default=None)

        option2: Union[int,str] = 289
        option3: Union[int,str] = Field(description="Description for option1", default="Salsa")

        # default
        id0: str = "Lala"
        id1: str = Field(description="Description for id2", default="Loop")
        id2: str = Field(default_factory=lambda: "Lamp")
        id3: list[str] = Field(default_factory=lambda: ["Lamp"])
        id4: list[int] = Field(default_factory=lambda: [4,7])

        cloth: str = field(default="A value")

    json_schema1, _ = build_root_json_schema(Person, "output")
    print(json_schema1)

    assert json_schema1 == {
        'properties': {'name': {'title': 'Name', 'type': 'string'}, 
                       'occupation': {'description': 'A peaceful occupation', 'title': 'Occupation', 'type': 'string'}, 
                       'age0': {'anyOf': [{'type': 'integer'}, {'type': 'null'}], 'title': 'Age0'}, 
                       'age1': {'anyOf': [{'type': 'integer'}, {'type': 'null'}], 'description': 'Description for age1', 'title': 'Age1'}, 
                       'option0': {'anyOf': [{'type': 'integer'}, {'type': 'string'}], 'title': 'Option0'}, 
                       'option1': {'anyOf': [{'type': 'integer'}, {'type': 'string'}], 'description': 'Description for option1', 'title': 'Option1'}, 
                       'age2': {'anyOf': [{'type': 'integer'}, {'type': 'null'}], 'default': None, 'title': 'Age2'}, 
                       'age3': {'anyOf': [{'type': 'integer'}, {'type': 'null'}], 'default': None, 'description': 'Description for age3', 'title': 'Age3'}, 
                       'option2': {'anyOf': [{'type': 'integer'}, {'type': 'string'}], 'default': 289, 'title': 'Option2'}, 
                       'option3': {'anyOf': [{'type': 'integer'}, {'type': 'string'}], 'default': 'Salsa', 'description': 'Description for option1', 'title': 'Option3'}, 
                       'id0': {'default': 'Lala', 'title': 'Id0', 'type': 'string'}, 
                       'id1': {'default': 'Loop', 'description': 'Description for id2', 'title': 'Id1', 'type': 'string'}, 
                       'id2': {'title': 'Id2', 'type': 'string'}, 
                       'id3': {'items': {'type': 'string'}, 'title': 'Id3', 'type': 'array'}, 
                       'id4': {'items': {'type': 'integer'}, 'title': 'Id4', 'type': 'array'}, 
                       'cloth': {'default': 'A value', 'title': 'Cloth', 'type': 'string'}},                        
                       'required': ['name', 'occupation', 'age0', 'age1', 'option0', 'option1'], 
                       'title': 'Person', 
                       'type': 'object'}


    json_schema2 = json_schema_massage(json_schema1)

    assert json_schema2 == {
        'properties': {'name': {'type': 'string'},
   'occupation': {'description': 'A peaceful occupation', 'type': 'string'},
   'age0': {'anyOf': [{'type': 'integer'}, {'type': 'null'}]},
   'age1': {'anyOf': [{'type': 'integer'}, {'type': 'null'}],
    'description': 'Description for age1'},
   'option0': {'anyOf': [{'type': 'integer'}, {'type': 'string'}]},
   'option1': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
    'description': 'Description for option1'},
   'age2': {'anyOf': [{'type': 'integer'}, {'type': 'null'}], 'default': None},
   'age3': {'anyOf': [{'type': 'integer'}, {'type': 'null'}],
    'description': 'Description for age3',
    'default': None},
   'option2': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
    'default': 289},
   'option3': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
    'description': 'Description for option1',
    'default': 'Salsa'},
   'id0': {'type': 'string', 'default': 'Lala'},
   'id1': {'description': 'Description for id2',
    'type': 'string',
    'default': 'Loop'},
   'id2': {'type': 'string'},
   'id3': {'items': {'type': 'string'}, 'type': 'array'},
   'id4': {'items': {'type': 'integer'}, 'type': 'array'},
   'cloth': {'type': 'string', 'default': 'A value'}},
  'required': ['name', 'occupation', 'age0', 'age1', 'option0', 'option1'],
  'type': 'object'}






def test_json_schema_pydantic2():

    class AAA(BaseModel):
        zoom: int
        zaam: str
        
    class Person(BaseModel):
        name: str  
        song: AAA
        occupation: str  
        location: str = "lala"
        age: Optional[AAA] = None
        unio: Union[int,str]

    json_schema1, _ = build_root_json_schema(Person, "output")
    print(json_schema1)

    assert json_schema1 == {'$defs': {'AAA': {'properties': {'zoom': {'title': 'Zoom',
      'type': 'integer'},
     'zaam': {'title': 'Zaam', 'type': 'string'}},
    'required': ['zoom', 'zaam'],
    'title': 'AAA',
    'type': 'object'}},
  'properties': {'name': {'title': 'Name', 'type': 'string'},
   'song': {'$ref': '#/$defs/AAA'},
   'occupation': {'title': 'Occupation', 'type': 'string'},
   'location': {'default': 'lala', 'title': 'Location', 'type': 'string'},
   'age': {'anyOf': [{'$ref': '#/$defs/AAA'}, {'type': 'null'}],
    'default': None},
   'unio': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
    'title': 'Unio'}},
  'required': ['name', 'song', 'occupation', 'unio'],
  'title': 'Person',
  'type': 'object'}


    json_schema2 = json_schema_massage(json_schema1)

    assert json_schema2 == {'properties': {'name': {'type': 'string'},
   'song': {'properties': {'zoom': {'type': 'integer'},
     'zaam': {'type': 'string'}},
    'required': ['zoom', 'zaam'],
    'type': 'object'},
   'occupation': {'type': 'string'},
   'location': {'type': 'string', 'default': 'lala'},
   'age': {'anyOf': [{'properties': {'zoom': {'type': 'integer'},
       'zaam': {'type': 'string'}},
      'required': ['zoom', 'zaam'],
      'type': 'object'},
     {'type': 'null'}],
    'default': None},
   'unio': {'anyOf': [{'type': 'integer'}, {'type': 'string'}]}},
  'required': ['name', 'song', 'occupation', 'unio'],
  'type': 'object'}





def test_json_schema_dataclass():

    @dataclass
    class Person:
        name: str  
        occupation: Annotated[str, "A peaceful occupation"]

        # Optional, Union
        age0: Optional[int]
        age1: Annotated[Optional[int], "Description for age1"]

        option0: Union[int,str]
        option1: Annotated[Union[int,str], "Description for option1"]

        # defaults start here
        age2: Optional[int] = None
        age3: Annotated[Optional[int], "Description for age3"] = None

        option2: Union[int,str] = 289
        option3: Annotated[Union[int,str], "Description for option1"] = "Salsa"

        # default
        id0: str = "Lala"
        id1: Annotated[str, "Description for id2"] = "Loop"
        id2: str = field(default_factory=lambda: "Lamp")
        id3: list[str] = field(default_factory=lambda: ["Lamp"])
        id4: list[int] = field(default_factory=lambda: [4,7])

        cloth: str = field(default="A value")


    json_schema1, _ = build_root_json_schema(Person, "output")

    assert json_schema1 == {
        'properties': {'name': {'type': 'string'},
    'occupation': {'type': 'string', 'description': 'A peaceful occupation'},
    'age0': {'anyOf': [{'type': 'integer'}, {'type': 'null'}]},
    'age1': {'anyOf': [{'type': 'integer'}, {'type': 'null'}],
        'description': 'Description for age1'},
    'option0': {'anyOf': [{'type': 'integer'}, {'type': 'string'}]},
    'option1': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
        'description': 'Description for option1'},
    'age2': {'anyOf': [{'type': 'integer'}, {'type': 'null'}], 'default': None},
    'age3': {'anyOf': [{'type': 'integer'}, {'type': 'null'}],
        'description': 'Description for age3',
        'default': None},
    'option2': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
        'default': 289},
    'option3': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
        'description': 'Description for option1',
        'default': 'Salsa'},
    'id0': {'type': 'string', 'default': 'Lala'},
    'id1': {'type': 'string',
        'description': 'Description for id2',
        'default': 'Loop'},
    'id2': {'type': 'string', 'default': 'Lamp'},
    'id3': {'items': {'type': 'string'}, 'type': 'array', 'default': ['Lamp']},
    'id4': {'items': {'type': 'integer'}, 'type': 'array', 'default': [4, 7]},
    'cloth': {'type': 'string', 'default': 'A value'}},
    'required': ['name', 'occupation', 'age0', 'age1', 'option0', 'option1'],
    'type': 'object'}

    json_schema2 = json_schema_massage(json_schema1)

    assert json_schema2 == {
        'properties': {'name': {'type': 'string'},
   'occupation': {'type': 'string', 'description': 'A peaceful occupation'},
   'age0': {'anyOf': [{'type': 'integer'}, {'type': 'null'}]},
   'age1': {'anyOf': [{'type': 'integer'}, {'type': 'null'}],
    'description': 'Description for age1'},
   'option0': {'anyOf': [{'type': 'integer'}, {'type': 'string'}]},
   'option1': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
    'description': 'Description for option1'},
   'age2': {'anyOf': [{'type': 'integer'}, {'type': 'null'}], 'default': None},
   'age3': {'anyOf': [{'type': 'integer'}, {'type': 'null'}],
    'description': 'Description for age3',
    'default': None},
   'option2': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
    'default': 289},
   'option3': {'anyOf': [{'type': 'integer'}, {'type': 'string'}],
    'description': 'Description for option1',
    'default': 'Salsa'},
   'id0': {'type': 'string', 'default': 'Lala'},
   'id1': {'type': 'string',
    'description': 'Description for id2',
    'default': 'Loop'},
   'id2': {'type': 'string', 'default': 'Lamp'},
   'id3': {'items': {'type': 'string'}, 'type': 'array', 'default': ['Lamp']},
   'id4': {'items': {'type': 'integer'}, 'type': 'array', 'default': [4, 7]},
   'cloth': {'type': 'string', 'default': 'A value'}},
  'required': ['name', 'occupation', 'age0', 'age1', 'option0', 'option1'],
  'type': 'object'}
    



def test_json_schema_dataclass2():

    @dataclass
    class AAA():  
        zoom: int
        zaam: str
        
    @dataclass
    class Person():  
        name: str  
        song: AAA
        occupation: str  
        unio: Union[int,str]
        koor: int

        location: str = "lala"
        age: Optional[AAA] = None
        age: AAA = field(default_factory=lambda: AAA(2,"two")) # = Sub(2)
        rage: list[int] = field(default_factory=lambda: [2,4]) # field(default_factory=list)
        location: str = "lala"

    json_schema1, _ = build_root_json_schema(Person, "output")


    assert json_schema1 == {'properties': {'name': {'type': 'string'},
   'song': {'properties': {'zoom': {'type': 'integer'},
     'zaam': {'type': 'string'}},
    'required': ['zoom', 'zaam'],
    'type': 'object'},
   'occupation': {'type': 'string'},
   'unio': {'anyOf': [{'type': 'integer'}, {'type': 'string'}]},
   'koor': {'type': 'integer'},
   'location': {'type': 'string', 'default': 'lala'},
   'age': {'properties': {'zoom': {'type': 'integer'},
     'zaam': {'type': 'string'}},
    'required': ['zoom', 'zaam'],
    'type': 'object',
    'default': {'zoom': 2, 'zaam': 'two'}},
   'rage': {'items': {'type': 'integer'}, 'type': 'array', 'default': [2, 4]}},
  'required': ['name', 'song', 'occupation', 'unio', 'koor'],
  'type': 'object'}

    json_schema2 = json_schema_massage(json_schema1)

    assert json_schema2 == {'properties': {'name': {'type': 'string'},
   'song': {'properties': {'zoom': {'type': 'integer'},
     'zaam': {'type': 'string'}},
    'required': ['zoom', 'zaam'],
    'type': 'object'},
   'occupation': {'type': 'string'},
   'unio': {'anyOf': [{'type': 'integer'}, {'type': 'string'}]},
   'koor': {'type': 'integer'},
   'location': {'type': 'string', 'default': 'lala'},
   'age': {'properties': {'zoom': {'type': 'integer'},
     'zaam': {'type': 'string'}},
    'required': ['zoom', 'zaam'],
    'type': 'object',
    'default': {'zoom': 2, 'zaam': 'two'}},
   'rage': {'items': {'type': 'integer'}, 'type': 'array', 'default': [2, 4]}},
  'required': ['name', 'song', 'occupation', 'unio', 'koor'],
  'type': 'object'}





"""Threads of messages are the way to communicate with models. 
For a quick extraction or classification, a thread with a simple IN message and possibly model instructions is enough.
For more sophisticated queries that depend on previous interactions, IN (user or input) and OUT (model output or response) messages can be added to construct a conversation or thread.

- Msg: A message of type IN (user query), OUT (assistant answer) or INST (initial model instructions).
- Thread: A sequence of Msgs alternating between user and model.
"""

from typing import Optional, Union, Callable
from typing_extensions import Self
from enum import Enum, IntFlag
from collections.abc import Sequence, Iterable
from dataclasses import dataclass

import json

import logging
logger = logging.getLogger(__name__)

from .utils import (
    quote_text,
    join_text,
    is_url,
    load_image_as_base64_data_url,
    download_as_base64_data_url
)


@dataclass
class Msg():

    class Kind(Enum):
        """Enumeration for kinds of messages in a Thread."""

        IN = "IN"
        """Input message, from user."""

        OUT = "OUT"
        """Model output message."""

        INST = "INST"
        """Initial model instructions."""
        

        def as_chatml_role(self: Self) -> str:
            CHATML_FROM_KIND: dict = {Msg.Kind.IN: "user", Msg.Kind.OUT: "assistant", Msg.Kind.INST: "system"}
            return CHATML_FROM_KIND.get(self) # type: ignore[return-value]
                
        @staticmethod
        def from_chatml_role(role: str) -> 'Msg.Kind':
            KIND_FROM_CHATML: dict = {"user": Msg.Kind.IN, "assistant": Msg.Kind.OUT, "system": Msg.Kind.INST}
            kind = KIND_FROM_CHATML.get(role)
            if kind is None:
                raise ValueError(f"Unknown ChatML role '{role}'")
            else:
                return kind
    
        @staticmethod
        def flip(kind: 'Msg.Kind') -> 'Msg.Kind':
            return Msg.Kind.OUT if kind is Msg.Kind.IN else Msg.Kind.IN
    
        def __repr__(self):
            return repr(self.value)
        
        

    kind: Kind
    """Message kind."""

    text: str
    """Message text (mandatory)."""

    images: Optional[list[dict]] = None
    """List of images in message. An entry must have a 'url' key, but any other keys can be added.
    Key 'url' key must be a remote url (https,http) or a 'data:' base64-encoded url."""
    


    def __post_init__(self):
        self.set_images(self.images)


    @staticmethod
    def make_IN(text: str,
                images: Optional[Union[list,str,dict]] = None) -> 'Msg':
        return Msg(Msg.Kind.IN,
                   text,
                   images)

    @staticmethod
    def make_OUT(text: str,
                 images: Optional[Union[list,str,dict]] = None) -> 'Msg':
        return Msg(Msg.Kind.OUT, 
                   text, 
                   images)

    @staticmethod
    def make_INST(text: str,
                  images: Optional[Union[list,str,dict]] = None) -> 'Msg':
        return Msg(Msg.Kind.INST, 
                   text,
                   images)


    def clone(self) -> 'Msg':
        return Msg(self.kind, self.text, self.images)


    def as_dict(self) -> dict:
        """Return Msg as a dict."""
        return {"kind": self.kind.value, # kind as string
                "text": self.text,
                "images": self.images}

    @staticmethod
    def from_dict(dic: dict) -> 'Msg':
        return Msg(kind=Msg.Kind(dic["kind"]),
                   text=dic["text"],
                   images=dic["images"])


    
    def as_chatml(self) -> dict:
        """Returns message in a ChatML dict.

        Returns:
            A ChatML dict with "role" and "content" keys.
        """

        role = self.kind.as_chatml_role()

        if self.images:
            chatml_msg = {
                "role": role, 
                "content": [
                    {"type": "text", "text": self.text},
                ]}
            
            for image in self.images:
                if "url" not in image:
                    raise ValueError(f"Image without 'url' key at {image}")
                
                image_url = {"url": image["url"]}
                if "detail" in image:
                    image_url["detail"] = image["detail"]

                chatml_msg["content"].append( # type: ignore[attr-defined]
                    {"type": "image_url", "image_url": image_url}
                )
            return chatml_msg
        else:
            return {"role": role, "content": self.text}


    @staticmethod
    def from_chatml(dic: dict,
                    join_sep:str = "\n") -> 'Msg':
        
        role = dic.get("role")
        if role is None:
            raise ValueError(f"Key 'role' not found in {dic}")

        kind = Msg.Kind.from_chatml_role(role)

        content = dic.get("content")
        if content is None:
            raise ValueError(f"Bad 'content' key in {dic}")

        text = ''
        images = []
        if isinstance(content, list):
            for cont in content:
                if not isinstance(cont, dict) or "type" not in cont:
                    raise TypeError(f"ChatML list entries must be of type dict and include a 'type' key in {cont}")

                if cont["type"] == "text":
                    text = join_text(text, cont["text"], join_sep)

                elif cont["type"] == "image_url":
                    image = cont["image_url"]
                    if "url" not in image:
                        raise TypeError(f"ChatML image_url entries must include a 'url' key in {cont}")
                    images.append(image)
                    
        elif isinstance(content, str):
            text = content
            
        else:
            raise TypeError(f"ChatML content must have str or dict type in {content}")

        return Msg(kind, 
                   text,
                   images if images else None)


    def set_images(self,
                   images: Union[list,str,dict,None]):

        if images is None:
            self.images = None
            return
        
        elif not isinstance(images, list): # ensure input images to be a list
            images = [images]

        self.images = []

        for image in images:

            if isinstance(image, str): # an URL or local path to be loaded into a data: URL
                if not is_url(image):
                    image_url = load_image_as_base64_data_url(image)
                else:
                    image_url = image

                self.images.append({"url": image_url})

            elif isinstance(image, dict):
                if "url" not in image:
                    raise TypeError(f"Image entries must have an 'url' key at {image}")
                
                if not is_url(image["url"]): # a local image path
                    image_url = load_image_as_base64_data_url(image["url"])
                    image["url"] = image_url
                    
                self.images.append(image)

            else:
                raise ValueError("Unable to set images: expecting an str file path or a dict with 'url' key")


    @property
    def has_images(self) -> bool:
        return bool(self.images)


    def download_images_as_data(self):
        """Download any remote images to a 'data:' url."""

        if self.images:
            for image in self.images:
                if not image["url"].startswith("data:"):
                    image["url"] = download_as_base64_data_url(image["url"])


    def join_same_kind(self,
                       other: Self,
                       join_sep: str):
        
        assert self.kind == other.kind, f"Messages are not of the same kind: {self} and {other}."

        self.text = join_text(self.text, other.text,
                              join_sep)
        
        if other.images:
            if self.images is None:
                self.images = []

            self.images += other.images


    def __str__(self):
        text = self.text.replace("\n", "\\n")
        out = f"{self.kind.name}={quote_text(text)}"
        if self.images:
            out += " images=["
            for image in self.images:
                dic = image.copy()
                dic['url'] = dic['url'][:64] + "(...)" if len(dic['url']) > 64 else dic['url']
                out += str(dic) + ","
            out += "]"
                
        return out

    def __repr__(self):
        return self.__str__()
    







class Thread(Sequence):
    """A sequence of messages alternating between IN ("user" role) and OUT ("assistant" role).

    Stores a special initial INST information (known as "system" role in ChatML) providing instructions to the model.
    Some models don't use system instructions - in those cases it's prepended to first IN message.

    Messages are kept in a strict IN,OUT,IN,OUT,... order. 
    To enforce this, if two IN messages are added, the second just appends to the text of the first or to its image list.
    
    Attributes:
        inst: Text for system instructions.
    """

    _msgs: list[Msg]
    """List of IN and OUT Msg messages. Messages are made to alternate between IN and OUT kinds."""

    inst: Msg
    """System instructions in an Msg of kind INST, defaults to empty text."""

    join_sep: str
    """Separator used when message text needs to be joined. Defaults to '\\n'"""


    def __init__(self,
                 t: Optional[Union[Self,list,Msg,dict,tuple,str]] = None,
                 inst: str = "",
                 join_sep: str = "\n"):
        """
        Args:
            t: Optionally initialize from a Thread, list[Msg], list[ChatML format dict], list[tuple], list[str], Msg, ChatML format dict, tuple or str.
            inst: Instructions text. If inst arg is not set and t is a Thread, its inst will be used.
            join_sep: Separator used when message text needs to be joined. Defaults to "\\n".

        Raises:
            TypeError: On invalid args passed.
        """

        self._msgs = []
        self.inst = Msg.make_INST(inst)
        self.join_sep = join_sep

        if t is not None:
            self.concat(t)



    def clone(self) -> Self:
        """Return a copy of current Thread.

        Returns:
            A copy of this Thread.
        """
        return Thread(self)


    def clear(self,
              clear_inst: bool = True):
        """Delete all messages and clear inst."""
        self._msgs = []
        if clear_inst:
            self.inst.text = ""
                


    # = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = Msg add and concat:
    
    def init_INST_IN(self,
                     inst_text: str,
                     in_text: str,
                     in_images: Optional[Union[list,str,dict]] = None):
        """Initialize Thread with instructions and an IN message.

        Args:
            inst_text: Instructions text.
            in_text: Text for IN message.
            in_images: An array (or its first element) of either an str (a file path, will be loaded and converted to a data: URL) or a dict with "url" key and others. If url arg is not a valid URL, it will be loaded and converted to a data: URL.
        """
        self.clear()
        self.inst.text = inst_text
        self.add_IN(in_text, in_images)


    def add_IN(self,
               in_text: str,
               in_images: Optional[Union[list,str,dict]] = None):
        """Appends an IN message to Thread.

        Args:
            in_text: Text for IN message.
            in_images: An array (or its first element) of either an str (a file path, will be loaded and converted to a data: URL) or a dict with "url" key and others. If url arg is not a valid URL, it will be loaded and converted to a data: URL.
        """
        self.add(Msg.Kind.IN, in_text, in_images)
    

    def add_OUT(self,
                out_text: str,
                out_images: Optional[Union[list,str,dict]] = None):
        """Appends an OUT message to Thread.

        Args:
            out_text: Text for OUT message.
            out_images: An array (or its first element) of either an str (a file path, will be loaded and converted to a data: URL) or a dict with "url" key and others. If url arg is not a valid URL, it will be loaded and converted to a data: URL.
        """
        self.add(Msg.Kind.OUT, out_text, out_images)

    
    def add_OUT_IN(self,
                   out_text: str,
                   in_text: str,
                   *,
                   out_images: Optional[Union[list,str,dict]] = None,
                   in_images: Optional[Union[list,str,dict]] = None):
        """Appends an OUT message followed by an IN message.

        Args:
            out_text: Text for OUT message.
            in_text: Text for IN message.
            out_images: An array (or its first element) of either an str (a file path, will be loaded and converted to a data: URL) or a dict with "url" key and others. If url arg is not a valid URL, it will be loaded and converted to a data: URL.
            in_images: Optional list of IN message images.
        """        
        self.add(Msg.Kind.OUT, out_text, out_images)
        self.add(Msg.Kind.IN, in_text, in_images)



    @staticmethod
    def make_INST_IN(inst_text: str,
                     in_text: str,
                     in_images: Optional[Union[list,str,dict]] = None) -> 'Thread':
        """Return an initialized Thread with instructions and an IN message.

        Args:
            inst_text: Instructions text.
            in_text: Text for IN message.
            in_images: An array (or its first element) of either an str (a file path, will be loaded and converted to a data: URL) or a dict with "url" key and others. If url arg is not a valid URL, it will be loaded and converted to a data: URL.
        """

        thread = Thread(inst=inst_text)
        thread.add_IN(in_text, in_images)
        return thread

    @staticmethod
    def make_IN(in_text: str,
                in_images: Optional[Union[list,str,dict]] = None) -> 'Thread':
        """Return an initialized Thread with an IN message.

        Args:
            in_text: Text for IN message.
            in_images: An array (or its first element) of either an str (a file path, will be loaded and converted to a data: URL) or a dict with "url" key and others. If url arg is not a valid URL, it will be loaded and converted to a data: URL.
        """

        thread = Thread()
        thread.add_IN(in_text, in_images)
        return thread



    def add(self, 
            t: Union[Msg,dict,tuple,str,Msg.Kind],
            text: Optional[str] = None,
            images: Optional[Union[list,str,dict]] = None):
        
        """Add a message to Thread.

        Accepts any of these argument combinations:
            t=Msg, ChatML format dict, tuple or str
            --or--
            t=kind, text[, images]

        Args:
            t: One of Msg, ChatML format dict, tuple or str, or Msg.Kind.
            text: Message text, only if t=Msg.Kind.
            images: only if t=Msg.Kind or t=str-> an array (or its first element) of either an str (a file path, will be loaded and converted to a data: URL) or a dict with keys "url" and any other keys like "detail". If url arg is not a valid URL, it will be loaded and converted to a data URL.
        """

        if text is not None:
            if not isinstance(t, Msg.Kind):
                raise TypeError("When arg 'text' is given, first arg must be of type Msg.Kind")

            msg = Msg(t, text, images)

        else: # add from t arg
            if isinstance(t, dict): # ChatML formatted dict
                msg = Msg.from_chatml(t)


            elif isinstance(t, tuple):
                msg = Msg(self.next_kind,
                          *t)

            elif isinstance(t, str): # simple text
                msg = Msg(self.next_kind,
                          t,
                          images)
                
            elif isinstance(t, Msg):
                msg = t.clone()

            else:
                raise TypeError("Arg 't' must be one of: Msg, ChatML format dict, tuple or str")

           
        # now append to list
        if msg.kind == Msg.Kind.INST:
            self.inst.join_same_kind(msg, self.join_sep)

        else:
            if not len(self._msgs) or msg.kind == self.next_kind: # next different kind or empty
                self._msgs.append(msg)
            else: # new msg is of same kind as last existing message: join/append to it
                last = self._msgs[-1]
                last.join_same_kind(msg, self.join_sep)



    def concat(self,
               t: Union[Self,list,Msg,dict,tuple,str]):
        """Concatenate to current Thread: another Thread, list[Msg], list[ChatML format dict], list[str], Msg, ChatML format dict or str.

        if last message in self is the same kind of first in t, their text, images, etc will be joined.

        Args:
            t: A Thread, list[Msg], list[ChatML format dict], list[str], Msg, ChatML format dict or str.
        """
        if isinstance(t, Thread):
            for msg in t:
                self.add(msg)

            self.inst.join_same_kind(t.inst, self.join_sep)

        else:
            if not isinstance(t, list):
                t = [t]
            for msg in t:
                self.add(msg)



    # = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = To and from dict and ChatML, load-save

    @staticmethod
    def from_dict(state: dict) -> 'Thread':
        """Deserialize a Thread from a dict."""

        th = Thread()
        for dic in state["_msgs"]:
            th.add(Msg.from_dict(dic))
        th.inst = Msg.from_dict(state["inst"])
        th.join_sep = state["join_sep"]

        return th

    def as_dict(self) -> dict:
        """Serialize this Thread to a dict."""
 
        state = {"_msgs": [],
                 "inst": self.inst.as_dict(),
                 "join_sep": self.join_sep}
    
        for msg in self._msgs:
            state["_msgs"].append(msg.as_dict()) # type: ignore[attr-defined]

        return state


    
    def load(self,
             path: str,
             clear: bool):
        """Load this Thread from a JSON file.

        Args:
            path: Path of file to load.
            clear: Should thread be cleared of messages, including INST? If not will concatenate with existing ones.
        """

        with open(path, 'r', encoding='utf-8') as f:
            js = f.read()
        state = json.loads(js)

        if clear:
            self.clear()

        th = self.from_dict(state)
        self.concat(th)

    
    def save(self,
             path: str):
        """Serialize this Thread to a JSON file.

        Args:
            path: Path of file to save into.
        """
 
        state = self.as_dict()
    
        json_str = json.dumps(state, indent=2, default=vars)
    
        with open(path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    


    def as_chatml(self,
                  include_INST: bool = True) -> list[dict]:
        """Returns Thread as a list of ChatML messages.

        Returns:
            A list of ChatML dict elements with "role" and "content" keys.
        """
        msgs = []

        if self.inst.text and include_INST:
            msgs.append(self.inst.as_chatml())

        for msg in self._msgs:
            msgs.append(msg.as_chatml())
            
        return msgs




    # = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = Utils

    @property
    def has_images(self) -> bool:
        for msg in self.get_iter(True):
            if msg.has_images:
                return True
        return False


    def download_images_as_data(self):
        self.inst.download_images_as_data()

        for msg in self:
            if msg.has_images:
                msg.download_images_as_data()


    def get_iter(self,
                 include_set_inst: bool):
        """Return an iterator that can be used to cycle over messages.
        include_set_inst: If inst message is set, include it before all others.
        """
        class MsgIter:
            def __init__(self, 
                         thread: Thread,
                         include_inst: bool):
                self.thread = thread
                self.curr = -1 - int(include_inst)

            def __iter__(self):
                return self
            
            def __next__(self):
                self.curr += 1
                if self.curr == -1:
                    return self.thread.inst
                elif self.curr < len(self.thread):
                    return self.thread[self.curr]
                else:
                    raise StopIteration

        return MsgIter(self,
                       include_set_inst and bool(self.inst.text))


    @property
    def next_kind(self) -> Msg.Kind:
        """Get kind of next new message that can be added to thread .

        Returns:
            Kind of last message or Msg.Kind.IN if empty.
        """
        if not self._msgs: # empty
            return Msg.Kind.IN
        else:
            return Msg.Kind.flip(self._msgs[-1].kind)



    def has_text_lower(self,
                       text_lower: str) -> bool:
        """Can the lowercase text be found in one of the messages?

        Args:
            text_lower: The lowercase text to search for in messages.

        Returns:
            True if such text was found.
        """
        for msg in self._msgs:
            if text_lower in msg.text.lower():
                return True
            
        return False        
        


    @staticmethod
    def ensure(query: Union['Thread',Msg,tuple,str],
               inst: Optional[str] = None) -> 'Thread':
        """Utility to return a Thread from either a passed Thread or an str used as an IN message.

        Args:
            query: A Thread or a single IN message given as Msg, list, tuple or str. List and tuple should contain the same args as for creating Msg.
            inst: Instruction message for model. Will override Thread's inst. Defaults to None.

        Raises:
            TypeError: Arg query must be of type Thread or str.

        Returns:
            Initialized Thread object.
        """
        
        if isinstance(query, str):
            th = Thread.make_IN(query)
            if inst is not None:
                th.inst.text = inst

        elif isinstance(query, tuple):
            th = Thread.make_IN(*query)
            if inst is not None:
                th.inst.text = inst

        elif isinstance(query, Msg):
            if query.kind != Msg.Kind.IN:
                raise TypeError("Only 'IN' kind is allowed for Msg")
            th = Thread(query)
            if inst is not None:
                th.inst.text = inst

        elif isinstance(query, Thread):
            if inst is not None:
                th = Thread(query, inst) # a clone
            else:
                th = query

        else:
            raise TypeError("Arg query must be of type Thread, Msg, tuple or str")

        return th




    # = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = __lower_level__


    def __len__(self):
        return len(self._msgs)


    def __add__(self,
                other: Union[Self,list, Msg, dict, str]) -> Self:
        out = self.clone()
        out.concat(other)
        return out


    def __getitem__(self, # type: ignore[override]
                    index: int) -> Msg:
        
        if isinstance(index, int):
            if index < 0: # need to take care of negative index
                index = len(self._msgs) + index
            if index >= len(self._msgs):
                raise IndexError(f"Trying to access non-existent message at index {index}, allowed index: 0 to {len(self._msgs)-1}")

            return self._msgs[index]
        
        elif isinstance(index, slice):
            start, stop, step = index.indices(len(self._msgs)) # negative index are taken care here
            return self._msgs[slice(start, stop, step)]
        else:
            raise TypeError("Arg index must be int or slice types")


    def __delitem__(self, 
                    index: int):

        if not isinstance(index, int):
            raise TypeError("Arg index must be of int type")

        if index < 0: # need to take care of negative index
            index = len(self._msgs) + index
        if index >= len(self._msgs):
            raise IndexError(f"Trying to access non-existent message at index {index}, allowed index: 0 to {len(self._msgs)-1}")

        if index + 1 == len(self._msgs): # last: just delete
            del self._msgs[index]

        elif index == 0: # delete first message: this will leave an OUT message as first
            del self._msgs[0]
        
        else:
            # augment next of the same kind with previous (of the same kind) text
            # merge text and images
            if self._msgs[index - 1].kind != self._msgs[index + 1].kind:
                raise ValueError(f"Messages at index {index-1} and index {index+1} should have the same kind")
            
            self._msgs[index - 1].join_same_kind(self._msgs[index + 1],  self.join_sep)

            del self._msgs[index + 1] # delete next, which was joined to index-1
            del self._msgs[index] # delete requested


    def __iter__(self):
        # Default iterator doesn't include inst message.
        return self.get_iter(False)


    def __reversed__(self):
        return reversed(self._msgs)




    def __eq__(self, other) -> bool:
        return (self._msgs == other._msgs and 
                self.inst == other.inst and 
                self.join_sep == other.join_sep)


    def __str__(self):
        inst = self.inst.text.replace('\n', '\\n')
        join_sep = self.join_sep.replace('\n', '\\n')
        out = f"Thread inst={quote_text(inst)}, join_sep='{join_sep}', len={len(self._msgs)}"
        for index,msg in enumerate(self._msgs):
            out += f"\n{index}: {msg}"
               
        return out

    def __repr__(self):
        return self.__str__()
    





    class Trim(IntFlag):
        """Flags for Thread trimming."""

        NONE = 0
        """No trimming."""

        INST = 1
        """Can remove INST message."""

        IN = 2
        """Can remove IN messages."""

        OUT = 4
        """Can remove OUT messages."""
            
        KEEP_FIRST_IN = 1024
        """If trimming IN messages, never remove first one."""

        KEEP_FIRST_OUT = 2048
        """If trimming OUT messages, never remove first one."""


    def trim(self,
             trim_flags: Trim,
             max_token_len: int,
             thread_token_len_fn: Callable
             ) -> int:
        """Trim context by selectively removing older messages until thread fits max_token_len.

        Args:
            trim_flags: Flags to guide selection of which messages to remove.
            max_token_len: Cut messages until size is lower than this number. Defaults to None.
            thread_token_len_fn: A function that returns token count for a passed Thread.

        Example of a thread_token_len_fn that counts 1 char = 1 token:
            def thread_token_len_fn(thread: Thread) -> int:
                total = len(thread.inst.text)
                for msg in thread:
                    total += len(msg.text)
                    if msg.images:
                        total += len(str(msg.images))
                return total

        Returns:
            Trimming result: 1=trimmed messages to max_token_len, 0: no trimming was needed, -1: Unable to trim to max_token_len.
        """

        if trim_flags == Thread.Trim.NONE: # no trimming
            return 0
        
        thread = self.clone()

        any_trim = False
        
        while True:

            curr_len = thread_token_len_fn(thread)

            if curr_len <= max_token_len:
                break

            logger.debug(f"len={curr_len} / max={max_token_len}")

            if thread.inst.text and trim_flags & Thread.Trim.INST:
                thread.inst.text = ""
                any_trim = True
                logger.debug(f"Cutting INST {thread.inst.text[:40]}")
                continue

            # cut first possible message, starting from oldest first ones
            trimmed = False
            in_index = out_index = 0

            for index,msg in enumerate(thread):

                if msg.kind == Msg.Kind.IN:
                    if trim_flags & Thread.Trim.IN:
                        if not (trim_flags & Thread.Trim.KEEP_FIRST_IN and in_index == 0):
                            del thread[index]
                            trimmed = True
                            logger.debug(f"Cutting IN {msg.text[:40]}")
                            break
                    in_index += 1

                elif msg.kind == Msg.Kind.OUT:
                    if trim_flags & Thread.Trim.OUT:                        
                        if not (trim_flags & Thread.Trim.KEEP_FIRST_OUT and out_index == 0):
                            del thread[index]
                            trimmed = True
                            logger.debug(f"Cutting OUT {msg.text[:40]}")
                            break
                    out_index += 1
                
            if not trimmed:
                # all thread messages were cycled but not a single could be cut, so size remains the same
                # arriving here we did all we could for trim_flags but could not remove any more
                return -1
            else:
                any_trim = True

        # while end
        

        if any_trim:
            self._msgs = thread._msgs
            self.inst = thread.inst
            
        return int(any_trim)





import pytest

import os, subprocess, shutil, json
from typing import Any, Optional, Union, Literal, Annotated, get_origin, get_args

import logging
logging.basicConfig(level=logging.DEBUG)

from sibila import Models

from .utils import setup_env_models, teardown_env_models, setup_model, teardown_model
from .utils import run_cmd, run_text, run_json



DO_TEARDOWN = True

@pytest.fixture(autouse=True, scope="module")
def env():

    base_dir, models_dir = setup_env_models("cli", 
                                            change_cwd=True,
                                            full_clean=True)

    print("---> setup", base_dir)
    
    yield base_dir, models_dir

    # --------------------------- teardown
    if DO_TEARDOWN == False:
        return

    print("---> teardown", base_dir)

    teardown_env_models(base_dir)
    




def test_models(env):

    models_path = os.path.join(env[1], Models.MODELS_CONF_FILENAME)

    cmd = "sibila models -m models/ -s llamacpp:local-model local-model.gguf"
    run_json(cmd, 
             path=models_path,
             expected_json={
                    "llamacpp": {
                        "local-model": {
                            "name": "local-model.gguf"
                        }
                    }
                })


    cmd = "sibila models -m models/ -f llamacpp:local-model chatml"
    run_json(cmd, 
             path=models_path,
             expected_json={
                    "llamacpp": {
                        "local-model": {
                            "name": "local-model.gguf",
                            "format": "chatml"
                        }
                    }
                })


    cmd = "sibila models -m models/ -d llamacpp:local-model"
    run_cmd(cmd)


    cmd = "sibila models -m models/ -s llamacpp:local-model local-model.gguf chatml"
    run_json(cmd, 
             path=models_path,
             expected_json={
                    "llamacpp": {
                        "local-model": {
                            "name": "local-model.gguf",
                            "format": "chatml"
                        }
                    }
                })


    cmd = "sibila models -m models/ -sl llamacpp:local-model-link local-model"
    dic = run_json(cmd, 
                   path=models_path)
    assert ("llamacpp" in dic and 
            "local-model-link" in dic["llamacpp"] and 
            dic["llamacpp"]["local-model-link"] == "local-model")

    cmd = "sibila models -m models/ -l"
    res = run_text(cmd,
                   expected_text=[
                       "only local models",
                        "local-model",
                        "local-model.gguf",
                        "chatml",
                        "local-model-link"                       
                   ],
                   not_expected_text=["gpt-4"])


    cmd = "sibila models -m models/ -l local-model-link"
    res = run_text(cmd,
                   expected_text=[
                       "only local models",
                       "local-model-link",
                       "local-model"
                   ],
                   not_expected_text=["local-model.gguf",
                                      "chatml",
                                      "gpt-4"])


    cmd = "sibila models -m models/ -l -b"
    res = run_text(cmd,
                   expected_text=[
                       "local and base models",
                       "local-model",
                       "local-model.gguf",
                       "chatml",
                       "local-model-link",
                       "gpt-4"                       
                   ],
                   not_expected_text=[])


    cmd = "sibila models -m models/ -l local-model-link -b"
    res = run_text(cmd,
                   expected_text=[
                       "local and base models",
                       "local-model-link",
                       "local-model"
                   ],
                   not_expected_text=["local-model.gguf",
                                      "chatml",
                                      "gpt-4"])


    # Error: Cannot delete 'llamacpp:local-model', as entry 'llamacpp:local-model-link' links to it
    cmd = "sibila models -m models/ -d llamacpp:local-model"
    run_cmd(cmd, 
            expected_exit=1)


    cmd = "sibila models -m models/ -d llamacpp:local-model-link"
    run_cmd(cmd,
            expected_exit=0)

    cmd = "sibila models -m models/ -d llamacpp:local-model"
    run_json(cmd, 
             path=models_path,
             expected_json={})




def test_formats(env):

    formats_path = os.path.join(env[1], Models.FORMATS_CONF_FILENAME)

    cmd = """sibila formats -m models -s test-format "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}{% endfor %}{% if add_generation_prompt %}{{ '<|im_start|>assistant \n' }}{% endif %}" """
    run_json(cmd, 
             path=formats_path,
             expected_json={
                    "test-format": {
                        "template": "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}{% endfor %}{% if add_generation_prompt %}{{ '<|im_start|>assistant \n' }}{% endif %}"
                    }
                })


    cmd = """sibila formats -m models -s test-format-chatml chatml"""
    run_json(cmd, 
             path=formats_path,
             expected_json={
                    "test-format": {
                        "template": "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}{% endfor %}{% if add_generation_prompt %}{{ '<|im_start|>assistant \n' }}{% endif %}"
                    },
                    "test-format-chatml": {
                        "template": "chatml"
                    }
                })
    

    cmd = """sibila formats -m models -d test-format-chatml"""
    run_cmd(cmd,
            expected_exit=0)

    cmd = "sibila formats -m models -sl test-format-link test-format"
    run_json(cmd, 
             path=formats_path,
             expected_json={
                    "test-format": {
                        "template": "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}{% endfor %}{% if add_generation_prompt %}{{ '<|im_start|>assistant \n' }}{% endif %}"
                    },
                    "test-format-link": "test-format"
                })

    cmd = "sibila formats -m models -sl chatml-format-link chatml"
    run_json(cmd, 
             path=formats_path,
             expected_json={
                    "test-format": {
                        "template": "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}{% endfor %}{% if add_generation_prompt %}{{ '<|im_start|>assistant \n' }}{% endif %}"
                    },
                    "test-format-link": "test-format",
                    "chatml-format-link": "chatml"
                })


    cmd = "sibila formats -m models/ -l"
    res = run_text(cmd,
                   expected_text=[
                       "only local formats",
                       "test-format",
                       "{{",
                       "test-format-link",
                       "test-format",
                       "chatml-format-link"],
                   not_expected_text=["openchat"])


    cmd = "sibila formats -m models/ -l -b"
    res = run_text(cmd,
                   expected_text=[
                       "local and base formats",
                       "test-format",
                       "{{",
                       "test-format-link",
                       "test-format",
                       "chatml-format-link",
                       "openchat"],
                   not_expected_text=[])


    cmd = "sibila formats -m models/ -d chatml-format-link"
    run_cmd(cmd, 
            expected_exit=0)


    # Error: Cannot delete 'test-format', as entry 'test-format-link' links to it
    cmd = "sibila formats -m models/ -d test-format"
    run_cmd(cmd, 
            expected_exit=1)


    cmd = "sibila formats -m models/ -d test-format-link"
    run_cmd(cmd,
            expected_exit=0)

    cmd = "sibila formats -m models/ -d test-format"
    run_json(cmd, 
             path=formats_path,
             expected_json={})




def test_hub():

    cmd = "sibila hub -l openchat"
    res = run_text(cmd,
                   expected_text=[
                       "TheBloke/openchat-3.5-1210-GGUF",
                       "openchat-3.5-1210.Q4_K_M.gguf",
                       "openchat-3.5-1210.Q8_0.gguf",
                       "TheBloke/Seraph-openchat-3.5-1210-Slerp-GGUF",
                       "tsunemoto/openchat-3.5-1210-GGUF"],
                   not_expected_text=[])

    cmd = "sibila hub -l openchat -a TheBloke"
    res = run_text(cmd,
                   expected_text=[
                       "TheBloke/openchat-3.5-1210-GGUF",
                       "openchat-3.5-1210.Q4_K_M.gguf",
                       "openchat-3.5-1210.Q8_0.gguf",
                       "TheBloke/Seraph-openchat-3.5-1210-Slerp-GGUF"],
                   not_expected_text=["tsunemoto/openchat-3.5-1210-GGUF",
                                      "MaziyarPanahi/openchat"])
    
    cmd = "sibila hub -l openchat -a TheBloke -f q4_k_m"
    res = run_text(cmd,
                   expected_text=[
                       "TheBloke/openchat-3.5-1210-GGUF",
                       "openchat-3.5-1210.Q4_K_M.gguf",
                       "TheBloke/Seraph-openchat-3.5-1210-Slerp-GGUF"],
                   not_expected_text=["tsunemoto/openchat-3.5-1210-GGUF",
                                      "MaziyarPanahi/openchat",
                                      "openchat-3.5-1210.Q8_0.gguf"])

    cmd = """sibila hub -l "TheBloke/openchat 3.5 1210 GGUF" -a TheBloke -f q4_k_m"""
    res = run_text(cmd,
                   expected_text=[
                       "TheBloke/openchat-3.5-1210-GGUF",
                       "openchat-3.5-1210.Q4_K_M.gguf",
                       "TheBloke/Seraph-openchat-3.5-1210-Slerp-GGUF"],
                   not_expected_text=["tsunemoto/openchat-3.5-1210-GGUF",
                                      "MaziyarPanahi/openchat",
                                      "openchat-3.5-1210.Q8_0.gguf"])


    cmd = """sibila hub -l "TheBloke/openchat 3.5 1210 GGUF" -a TheBloke -f openchat-3.5-1210.Q4_K_M.gguf"""
    res = run_text(cmd,
                   expected_text=[
                       "TheBloke/openchat-3.5-1210-GGUF",
                       "openchat-3.5-1210.Q4_K_M.gguf"
                       ],
                   not_expected_text=["tsunemoto/openchat-3.5-1210-GGUF",
                                      "TheBloke/Seraph-openchat-3.5-1210-Slerp-GGUF",
                                      "MaziyarPanahi/openchat",
                                      "openchat-3.5-1210.Q8_0.gguf"])
   
    cmd = "sibila hub -l openchat -a TheBloke -f openchat-3.5-1210.Q4_K_M.gguf"
    res = run_text(cmd,
                   expected_text=[
                       "TheBloke/openchat-3.5-1210-GGUF",
                       "openchat-3.5-1210.Q4_K_M.gguf"
                       ],
                   not_expected_text=["tsunemoto/openchat-3.5-1210-GGUF",
                                      "TheBloke/Seraph-openchat-3.5-1210-Slerp-GGUF",
                                      "MaziyarPanahi/openchat",
                                      "openchat-3.5-1210.Q8_0.gguf"])





import pytest

import os, subprocess, shutil, json, asyncio
from typing import Any, Optional, Union, Literal, Annotated, get_origin, get_args

import logging
logger = logging.getLogger(__name__)
# pytest --log-cli-level=DEBUG

from sibila import (
    Models,
    LlamaCppModel,
    GenConf
)


from .utils import setup_env_models, teardown_env_models, setup_model, teardown_model
from .utils import run_cmd, run_text, run_json




# https://huggingface.co/TheBloke/TinyLlama-1.1B-Chat-v1.0-GGUF
MODEL_FILENAME = "tinyllama-1.1b-chat-v1.0.Q4_K_M.gguf"

DO_TEARDOWN = True

@pytest.fixture(autouse=True, scope="module")
def env_model():

    base_dir, models_dir = setup_env_models("llamacpp", 
                                            change_cwd=True,
                                            full_clean=False)

    model_path = setup_model(MODEL_FILENAME, 
                             models_dir,
                             always_copy=False)

    ret = base_dir, model_path
    print("---> setup", ret)
    
    yield ret

    # --------------------------- teardown
    if DO_TEARDOWN == False:
        return

    print("---> teardown", ret)

    teardown_model(model_path)

    teardown_env_models(base_dir)










def test_create_llamacppmodel(env_model):

    full_model_path = env_model[1]
    rel_model_path = os.path.join("models", MODEL_FILENAME)

    # model file not found
    with pytest.raises(NameError):
        model = LlamaCppModel(full_model_path + "NOT_THERE")

    # relative path
    model = LlamaCppModel(rel_model_path, format="zephyr")
    del model

    # absolute path
    model = LlamaCppModel(full_model_path)
    del model





def test_models(env_model):

    base_dir = env_model[0]
    models_dir_path = os.path.join(base_dir, "models")
    models_json_path = os.path.join(models_dir_path, "models.json")

    full_model_path = env_model[1]
    rel_model_path = os.path.join("models", MODEL_FILENAME)

    name = "local-model"
    res_name = f"llamacpp:{name}"
    link_name = f"{name}-link"
    link_res_name = f"llamacpp:{link_name}"


    # avoid any existing models.json
    if os.path.isfile(models_json_path):
        os.remove(models_json_path)

    # create model entry
    cmd = f"sibila models -m models/ -s {res_name} {MODEL_FILENAME} zephyr"
    run_json(cmd, 
             path=models_json_path,
             expected_json={
                    "llamacpp": {
                        name: {
                            "name": MODEL_FILENAME,
                            "format": "zephyr"
                        }
                    }
                })

    # create model link
    cmd = f"sibila models -m models/ -sl {link_res_name} {name}"
    run_json(cmd, 
             path=models_json_path,
             expected_json={
                    "llamacpp": {
                        name: {
                            "name": MODEL_FILENAME,
                            "format": "zephyr"
                        },
                        link_name: name
                    }
                })
    



    # model file not found
    with pytest.raises(NameError):
        model = Models.create(res_name + "NOT_THIS")

    # Models.setup() not called ===============================================
    with pytest.raises(NameError):
        model = Models.create(res_name)

    # rel_path
    model = Models.create("llamacpp:" + rel_model_path)
    del model

    # absolute path
    model = Models.create("llamacpp:" + full_model_path)
    del model


    # Models.setup() called ===================================================
    Models.setup("models")

    model = Models.create(res_name)
    del model

    model = Models.create(link_res_name)
    del model







def test_ctx_len(env_model):
    
    rel_model_path = os.path.join("models", MODEL_FILENAME)

    model = LlamaCppModel(rel_model_path, format="zephyr")
    assert model.ctx_len == 2048
    assert model.max_tokens_limit == 2048
    del model


    model = LlamaCppModel(rel_model_path, format="zephyr",
                          ctx_len=0)
    assert model.ctx_len == 2048
    assert model.max_tokens_limit == 2048
    del model

    model = LlamaCppModel(rel_model_path, format="zephyr",
                          ctx_len=1024)
    assert model.ctx_len == 1024
    assert model.max_tokens_limit == 1024
    del model






def test_max_tokens(env_model):
    
    rel_model_path = os.path.join("models", MODEL_FILENAME)

    model = LlamaCppModel(rel_model_path, format="zephyr")
    CTX_LEN = 2048

    assert model.calc_max_max_tokens(0) == CTX_LEN-0
    assert model.calc_max_max_tokens(500) == CTX_LEN-500
    assert model.calc_max_max_tokens(3000) == CTX_LEN-3000


    genconf=GenConf(max_tokens=1000)    
    assert genconf.resolve_max_tokens(model.ctx_len, model.max_tokens_limit) == 1000
    assert model.resolve_genconf_max_tokens(100, genconf) == 1000
    assert model.resolve_genconf_max_tokens(1500, genconf) == CTX_LEN - 1500

    genconf=GenConf(max_tokens=-20)
    assert genconf.resolve_max_tokens(model.ctx_len, model.max_tokens_limit) == int(CTX_LEN * 20/100)
    assert model.resolve_genconf_max_tokens(100, genconf) == int(CTX_LEN * 20/100)
    assert model.resolve_genconf_max_tokens(1900, genconf) == CTX_LEN - 1900

    genconf=GenConf(max_tokens=0)    
    assert genconf.resolve_max_tokens(model.ctx_len, model.max_tokens_limit) == 2048
    assert model.resolve_genconf_max_tokens(100, genconf) == CTX_LEN - 100
    assert model.resolve_genconf_max_tokens(1900, genconf) == CTX_LEN - 1900

    del model





def test_prompt(env_model):

    rel_model_path = os.path.join("models", MODEL_FILENAME)

    PROMPT = "Tell me about oranges"
    HAS = "orange"

    model = LlamaCppModel(rel_model_path, format="zephyr")
    text = model(PROMPT)
    assert HAS in text.lower()
    del model




INT_PROMPT = "extract a number from the following text: there are 12 bananas"
TRUE_PROMPT = "extract a True/False value from the following text: Yes I got it!"



def test_extract(env_model):

    rel_model_path = os.path.join("models", MODEL_FILENAME)

    model = LlamaCppModel(rel_model_path, format="zephyr")


    res = model.extract(int, INT_PROMPT)
    assert res == 12

    res = model.extract(bool, TRUE_PROMPT)
    assert res == True

    del model







def test_extract_async1(env_model):

    rel_model_path = os.path.join("models", MODEL_FILENAME)

    model = LlamaCppModel(rel_model_path, format="zephyr")

    async def run_async():
        print("run_async begin")

        res = await model.extract_async(int, INT_PROMPT)
        assert res == 12

        res = await model.extract_async(bool, TRUE_PROMPT)
        assert res == True
        print("run_async done")
        
    asyncio.run(run_async())



    async def run1_async():        
        res = await model.extract_async(int, INT_PROMPT)
        assert res == 12
        print("run1 done")

    async def run2_async():
        res = await model.extract_async(bool, TRUE_PROMPT)
        assert res == True
        print("run2 done")


    async def as_completed():
        print("as_complete begin")
        tasks = [run1_async(), run2_async()]
        for task in asyncio.as_completed(tasks):
            await task
        print("as_complete done")
            
    asyncio.run(as_completed())    



    async def gather():
        print("gather begin")
        tasks = [run1_async(), run2_async()]
        await asyncio.gather(*tasks)
        print("gather done")
            
    asyncio.run(gather())    


    del model






async def test_extract_async2(env_model):

    rel_model_path = os.path.join("models", MODEL_FILENAME)

    model = LlamaCppModel(rel_model_path, format="zephyr")

    print("run_async begin")

    res = await model.extract_async(int, INT_PROMPT)
    assert res == 12

    res = await model.extract_async(bool, TRUE_PROMPT)
    assert res == True
    print("run_async done")
        



"""OpenAI remote model access.
Using the chat completion API:
https://platform.openai.com/docs/api-reference/chat/create

- OpenAIModel: Access OpenAI models.
- OpenAITokenizer: Tokenizer for OpenAI models.
"""


from typing import Any, Optional, Union
import sys, json
from time import time 
from copy import copy

import logging
logger = logging.getLogger(__name__)


from .gen import (
    GenConf,
    GenOut
)

from .thread import (
    Thread,
    Msg
)

from .model import (
    MessagesModel,
    Tokenizer
)

from .json_schema import JSchemaConf

try:
    import openai
    has_openai = True
except ImportError:
    has_openai = False
    
try:
    import tiktoken
    has_tiktoken = True
except ImportError:
    has_tiktoken = False




class OpenAIModel(MessagesModel):
    """Access an OpenAI model.

    Supports constrained JSON output, via the OpenAI API tools mechanism.

    Ref:
        https://platform.openai.com/docs/api-reference/chat/create
    """

    PROVIDER_NAME:str = "openai"
    """Provider prefix that this class handles."""

    _token_estimation_factor: float
    """Multiplication factor to estimate token usage: multiplies text length to obtain token length."""

    DEFAULT_TOKEN_ESTIMATION_FACTOR: float = 0.4
    """Default factor for token_estimation_factor."""



    def __init__(self,
                 name: str,
                 *,
                 
                 # common base model args
                 genconf: Optional[GenConf] = None,
                 schemaconf: Optional[JSchemaConf] = None,
                 ctx_len: Optional[int] = None,
                 max_tokens_limit: Optional[int] = None,
                 tokenizer: Optional[Tokenizer] = None,
                
                 # most important OpenAI-specific args
                 api_key: Optional[str] = None,
                 base_url: Optional[str] = None,
                 overhead_per_msg: Optional[int] = None,
                 token_estimation_factor: Optional[float] = None,
                 create_tokenizer: bool = False,
                 
                 # other OpenAI-specific args
                 other_init_kwargs: dict = {},
                 ):
        """Create an OpenAI remote model.

        Args:
            name: Model name to resolve into an existing model.
            genconf: Model generation configuration. Defaults to None.
            schemaconf: Default configuration for JSON schema validation, used if generation call doesn't supply one. Defaults to None.
            ctx_len: Maximum context length to be used (shared for input and output). None for model's default.
            max_tokens_limit: Maximum output tokens limit. None for model's default.
            tokenizer: An external initialized tokenizer to use instead of the created from the GGUF file. Defaults to None.
            api_key: OpenAI API key. Defaults to None, which will use env variable OPENAI_API_KEY.
            base_url: Base location for API access. Defaults to None, which will use env variable OPENAI_BASE_URL or a default.
            overhead_per_msg: Overhead tokens to account for when calculating token length. None for model's default.
            token_estimation_factor: Used when no tokenizer is available. Multiplication factor to estimate token usage: multiplies total text length to obtain token length.
            create_tokenizer: When no tokenizer is passed, should try to create one?
            other_init_kwargs: Extra args for OpenAI.OpenAI() initialization. Defaults to {}.

        Raises:
            ImportError: If OpenAI API is not installed.
            NameError: If model name was not found or there's an API or authentication problem.
        """


        if not has_openai:
            raise ImportError("Please install openai by running: pip install openai")

        self._client = self._client_async = None


        # also accept "provider:name" for ease of use
        provider_name = self.PROVIDER_NAME + ":"
        if name.startswith(provider_name):
            name = name[len(provider_name):]

        super().__init__(False,
                         genconf,
                         schemaconf,
                         tokenizer
                         )

        if (ctx_len is not None and
            max_tokens_limit is not None and
            overhead_per_msg is not None and
            token_estimation_factor is not None): # all elements given: probably created via Models.create()

            self._model_name = name
            default_ctx_len = ctx_len
            default_max_tokens_limit = max_tokens_limit
            default_overhead_per_msg = overhead_per_msg
            default_token_estimation_factor = token_estimation_factor
        
        else: # need to resolve
            settings = self.resolve_settings(self.PROVIDER_NAME,
                                             name,
                                             ["name", 
                                              "ctx_len", 
                                              "max_tokens_limit", 
                                              "overhead_per_msg",
                                              "token_estimation_factor"])
            self._model_name = settings.get("name") or name
            default_ctx_len = settings.get("ctx_len") # type: ignore[assignment]
            default_max_tokens_limit = settings.get("max_tokens_limit") # type: ignore[assignment]
            default_overhead_per_msg = settings.get("overhead_per_msg") # type: ignore[assignment]
            default_token_estimation_factor = settings.get("token_estimation_factor") # type: ignore[assignment]
            
            # all defaults are conservative values
            if ctx_len is None and default_ctx_len is None:
                default_ctx_len = 4096
                logger.warning(f"Model '{self._model_name}': unknown ctx_len, assuming {default_ctx_len}")

            if max_tokens_limit is None and default_max_tokens_limit is None:
                default_max_tokens_limit = ctx_len or default_ctx_len                
                # don't warn: assume equal to ctx_len: logger.warning(f"Model '{self._model_name}': unknown max_tokens_limit, assuming {default_max_tokens_limit}")

            if overhead_per_msg is None and default_overhead_per_msg is None:
                default_overhead_per_msg = 3
                # don't warn for this setting due to derived model classes (none uses it)

            if token_estimation_factor is None and default_token_estimation_factor is None:
                default_token_estimation_factor = self.DEFAULT_TOKEN_ESTIMATION_FACTOR
                logger.warning(f"Model '{self._model_name}': unknown token_estimation_factor, assuming {default_token_estimation_factor}")


        self.ctx_len = ctx_len or default_ctx_len
        
        self.max_tokens_limit = max_tokens_limit or default_max_tokens_limit
        self.max_tokens_limit = min(self.max_tokens_limit, self.ctx_len)

        self._overhead_per_msg = overhead_per_msg or default_overhead_per_msg

        self._token_estimation_factor = token_estimation_factor or default_token_estimation_factor

        self.maybe_image_input = True # True means maybe - always check model specs

        # only check for "json" text presence as json schema (including field descriptions) is requested with the tools facility.
        self.json_format_instructors["json_schema"] = self.json_format_instructors["json"]


        if self.tokenizer is None and create_tokenizer:
            try:
                self.tokenizer = OpenAITokenizer(self._model_name)
            except Exception as e:
                logger.warning(f"Could not create a local tokenizer for model '{self._model_name}' - "
                               "token length calculation will be disabled and assume defaults. "
                               "To support recent OpenAI models, install the latest tiktoken version with 'pip install -U tiktoken'. "
                               f"Internal error: {e}")


        self._client_init_kwargs = other_init_kwargs
        if api_key is not None:
            self._client_init_kwargs["api_key"] = api_key
        if base_url is not None:
            self._client_init_kwargs["base_url"] = base_url



    def close(self):
        """Close model, release resources like memory or net connections."""
        self._client = self._client_async = self.tokenizer = None



    def _ensure_client(self,
                       is_async: bool):
        if is_async and self._client_async is not None:
            return
        elif not is_async and self._client is not None:
            return
            
        try:
            logger.debug(f"Creating inner OpenAI with ctx_len={self.ctx_len}, max_tokens_limit={self.max_tokens_limit}, "
                         f"_overhead_per_msg={self._overhead_per_msg}, _token_estimation_factor={self._token_estimation_factor}, "
                         f"init_kwargs={self._client_init_kwargs}")

            if is_async:
                self._client_async = openai.AsyncOpenAI(**self._client_init_kwargs) # type: ignore[assignment]
            else:
                self._client = openai.OpenAI(**self._client_init_kwargs) # type: ignore[assignment]

        except Exception as e:
            raise NameError(f"Could not create {'async' if is_async else ''} model '{self._model_name}' with error: {e}")
                








    def _gen_pre(self, 
                 thread: Thread,
                 genconf: Union[GenConf, None]
                 ) -> tuple:

        if genconf is None:
            genconf = self.genconf

        thread = self._prepare_gen_thread(thread, genconf)

        # This provider doesn't require "max_tokens" but will error on excess.
        if genconf.max_tokens != 0:
            token_len = self.token_len(thread, genconf)
            resolved_max_tokens = self.resolve_genconf_max_tokens(token_len, genconf)
            # for reference: next commented-out line is a bad idea, as endpoint will error if max_tokens is greater than limit:
            # resolved_max_tokens = self.resolve_genconf_max_tokens(0, genconf)
        else: # for genconf.max_tokens=0, this provider doesn't require "max_tokens", so we don't send below
            resolved_max_tokens = 0


        json_kwargs: dict = {}
        if genconf.format == "json":
            
            json_kwargs["response_format"] = {"type": "json_object"}

            if genconf.json_schema is not None:
                # use json_schema in OpenAi's tool API
                json_kwargs["tool_choice"] = {
                    "type": "function",
                    "function": {"name": self.output_fn_name},
                }

                if isinstance(genconf.json_schema, str):
                    params = json.loads(genconf.json_schema)
                else:
                    params = genconf.json_schema
                
                json_kwargs["tools"] = [
                    {
                        "type": "function",
                        "function": {
                            "name": self.output_fn_name,
                            "parameters": params
                        }
                    }
                ]

        # seed config is disabled, remote models and some hardware accelerated local models don't support it.
        # seed = genconf.seed
        # if seed == -1:
        #    seed = int(time())
        #    logger.debug(f"OpenAI random seed={seed}")

        
        msgs = thread.as_chatml()

        kwargs = {"model": self._model_name,
                  "messages": msgs, # type: ignore[arg-type]                  
                  "temperature": genconf.temperature,
                  "top_p": genconf.top_p,
                  # "seed": seed,
                  "n": 1,
                  **json_kwargs}
        
        if genconf.stop: # OpenAI API: empty stop errors when generating from images
            kwargs["stop"] = genconf.stop

        if resolved_max_tokens:
            kwargs["max_tokens"] = resolved_max_tokens

        # inject model-specific args, if any
        kwargs.update(genconf.resolve_special(self.PROVIDER_NAME))

        logger.debug(f"{type(self).__name__} gen args: {kwargs}")

        return (kwargs, genconf)
        

    def _gen_post(self, 
                  response: Any,
                  pre_kwargs: dict,
                  genconf: GenConf
                  ) -> GenOut:
            
        logger.debug(f"OpenAI response: {response}")

        choice = response.choices[0]
        finish = choice.finish_reason
        # OpenAI-compatible provider endpoints can give non-standard finish_reason values. Map as needed:
        if finish in ["eos", "tool_calls"]: finish = "stop"
        message = choice.message

        if "tool_choice" in pre_kwargs:
            
            # json schema generation via the tools API:
            if message.tool_calls is not None:
                if len(message.tool_calls) != 1:
                    logger.warn(f"OpenAIModel: expecting single message.tool_calls, but received {len(message.tool_calls)} - using first.")

                fn = message.tool_calls[0].function
                if fn.name != self.output_fn_name:
                    logger.warn(f"OpenAIModel: expecting '{self.output_fn_name}' function name, received ({fn.name})")

                text = fn.arguments

            else: # use content instead
                logger.warn("OpenAIModel: expecting message.tool_calls, but none received - using text content")
                text = message.content # type: ignore[assignment]
        
        else:
            # text or simple json format
            text = message.content # type: ignore[assignment]

        return self._prepare_gen_out(text, finish, genconf)







    
    def gen(self, 
            thread: Thread,
            genconf: Optional[GenConf] = None,
            ) -> GenOut:
        """Text generation from a Thread, used by the other model generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with result, generated text, etc.
            The output text is in GenOut.text.
        """

        genconf2: GenConf
        kwargs, genconf2 = self._gen_pre(thread, genconf)

        self._ensure_client(False)

        try:
            # https://platform.openai.com/docs/api-reference/chat/create
            response = self._client.chat.completions.create(**kwargs) # type: ignore[attr-defined]

        except Exception as e:
            raise RuntimeError(f"Cannot generate. Internal error: {e}")


        return self._gen_post(response,
                              kwargs,
                              genconf2)
    



    async def gen_async(self, 
                        thread: Thread,
                        genconf: Optional[GenConf] = None,
                        ) -> GenOut:
        """Async text generation from a Thread, used by the other (async) generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None.

        Raises:
            RuntimeError: If unable to generate.

        Returns:
            A GenOut object with result, generated text, etc.
            The output text is in GenOut.text.
        """

        genconf2: GenConf
        kwargs, genconf2 = self._gen_pre(thread, genconf)

        self._ensure_client(True)

        try:
            # https://platform.openai.com/docs/api-reference/chat/create
            response = await self._client_async.chat.completions.create(**kwargs) # type: ignore[attr-defined]

        except Exception as e:
            raise RuntimeError(f"Cannot generate. Internal error: {e}")

        return self._gen_post(response,
                              kwargs,
                              genconf2)
    

   





    def token_len(self,
                  thread_or_text: Union[Thread,str],
                  genconf: Optional[GenConf] = None) -> int:
        """Calculate or estimate the token length for a Thread or a plain text string.
        In some cases where it's not possible to calculate the exact token count, 
        this function should give a conservative (upper bound) estimate.
        It's up to the implementation whether to account for side information like JSON Schema,
        but it must reflect the model's context token accounting.
        Thread or text must be the final text which will passed to model.

        If a json_schema is provided in genconf, we use its string's token_len as upper bound for the extra prompt tokens.
        
        From https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
        
        More info on calculating function_call (and tools?) tokens:

        https://community.openai.com/t/how-to-calculate-the-tokens-when-using-function-call/266573/24
        
        https://gist.github.com/CGamesPlay/dd4f108f27e2eec145eedf5c717318f5

        Args:
            thread_or_text: For token length calculation.
            genconf: Model generation configuration. Defaults to None.

        Returns:
            Estimated number of tokens used.
        """

        if isinstance(thread_or_text, Thread):
            thread = thread_or_text            
        else:
            thread = Thread.make_IN(thread_or_text)

        num_tokens = 0

        if self.tokenizer is None: # no tokenizer was found, so we'll have to do a conservative estimate

            OVERHEAD_PER_MSG = 3
            for msg in thread.get_iter(True): # True for system message
                message = msg.as_chatml()
                msg_tokens = len(str(message["content"])) * self._token_estimation_factor + OVERHEAD_PER_MSG
                # str(message["content"]): hacky way to deal with dict "content" key
                num_tokens += int(msg_tokens)

            if genconf is not None and genconf.json_schema is not None:
                if isinstance(genconf.json_schema, str):
                    js_str = genconf.json_schema
                else:
                    js_str = json.dumps(genconf.json_schema)

                tools_num_tokens = len(js_str) * self._token_estimation_factor
                num_tokens += int(tools_num_tokens)
                # print("tools_num_tokens", tools_num_tokens)

        else: # do an "informed" token estimation from what is known of the OpenAI model's tokenization
            
            for msg in thread.get_iter(True): # True for system message
                message = msg.as_chatml()
                # print(message)
                num_tokens += self._overhead_per_msg
                for key, value in message.items():
                    num_tokens += len(self.tokenizer.encode(str(value))) # str(value): hacky way to deal with dict "content" key
            
            # add extras + every reply is primed with <|start|>assistant<|message|>
            num_tokens += 32

            # print("text token_len", num_tokens)

            if genconf is not None and genconf.json_schema is not None:
                TOOLS_TOKEN_LEN_FACTOR = 1.2

                if isinstance(genconf.json_schema, str):
                    js_str = genconf.json_schema
                else:
                    js_str = json.dumps(genconf.json_schema)

                tools_num_tokens = self.tokenizer.token_len(js_str)

                # this is an upper bound, as empirically tested with the api.
                tools_num_tokens = int(tools_num_tokens * TOOLS_TOKEN_LEN_FACTOR)
                # print("tools token_len", tools_num_tokens)

                num_tokens += tools_num_tokens

        
        return num_tokens



    @classmethod
    def known_models(cls,
                     api_key: Optional[str] = None) -> Union[list[str], None]:
        """List of model names that can be used. Some of the models are not chat models and cannot be used,
        for example embedding models.
        
        Args:
            api_key: Requires OpenAI API key, passed as this arg or set in env variable OPENAI_API_KEY.

        Returns:
            Returns a list of known models.
        """
 
        client = openai.OpenAI(api_key=api_key)
        model_list = client.models.list()

        out = []
        for model in model_list.data:
            out.append(model.id)
        return sorted(out)





    def name(self) -> str:
        """Model (short) name."""
        return self._model_name
    
    def desc(self) -> str:
        """Model description."""
        return f"{type(self).__name__}: '{self._model_name}'"


    @classmethod
    def provider_version(_) -> str:
        """Provider library version: provider x.y.z
        Ex. openai-1.3.6
        """
        try:        
            import openai
            ver = openai.__version__
        except Exception:
            raise ImportError("Please install openai by running: pip install openai")
            
        return f"openai-{ver}"
    
    








class OpenAITokenizer(Tokenizer):
    """Tokenizer for OpenAI models."""

    def __init__(self, 
                 model: str
                 ):

        if not has_tiktoken:
            raise Exception("Please install tiktoken by running: pip install tiktoken")
        
        self._tok = tiktoken.encoding_for_model(model)

        self.vocab_size = self._tok.n_vocab
        
        self.bos_token_id = None
        self.bos_token = None

        self.eos_token_id = None
        self.eos_token = None

        self.pad_token_id = None
        self.pad_token = None

        self.unk_token_id = None
        self.unk_token = None
  

    
    def encode(self, 
               text: str) -> list[int]:
        """Encode text into model tokens. Inverse of Decode().

        Args:
            text: Text to be encoded.

        Returns:
            A list of ints with the encoded tokens.
        """
        return self._tok.encode(text)

        
    def decode(self, 
               token_ids: list[int],
               skip_special: bool = True) -> str:
        """Decode model tokens to text. Inverse of Encode().

        Args:
            token_ids: List of model tokens.
            skip_special: Don't decode special tokens like bos and eos. Defaults to True.

        Returns:
            Decoded text.
        """
        assert skip_special, "OpenAITokenizer only supports skip_special=True"

        return self._tok.decode(token_ids)

   




